#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you should cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.

# File   : peptml.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 22.06.2020


import  textwrap

import  numpy   as      np
from    tqdm    import  tqdm

import  pept

from    pept.tracking                           import  peptml
import  pept.tracking.trajectory_separation     as      tsp

from    pept.visualisation                      import  PlotlyGrapher


# Colours for printing to the terminal, taken from blender build scripts.
class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'        # Return to normal colour
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


class PEPTMLUser:
    '''This is an example script analysing PEPT data using the PEPT-ML
    algorithm from the `pept` library.

    The script is encapsulated in a class such that, after running it (i.e.
    instantiating the class), all the relevant analysis objects can be accessed
    as class attributes (e.g. LoRs, clusterers, tracer locations).

    The class has three methods:

    1. __init__ : the class constructor that is run when instantiating it; it
       contains the actual PEPT-ML algorithm.
    2. typecheck_and_create_attributes : typecheck input parameters and create
       class attributes.
    3. minimal : the minimal PEPT-ML code, without pretty-printing to the
       terminal.

    It is more than recommended to look through the code, copy and hack into
    it! The class constructor (`__init__`) contains the PEPT-ML algorithm with
    printing functionality to the terminal; if you'd like the code without such
    additions, use the equivalent code in the `minimal` method.

    The main steps of the PEPT-ML-User algorithm are:

    -- Done beforehand, by the user --

    1. Initialise LoRs in a `pept.LineData` instance (if they are stored in a
       file, you can use `pept.utilities.read_csv`). **You must do this before
       using this script**.

    -- Done by this script --

    2. Compute the cutpoints from the LoRs using the `peptml.Cutpoints` class.
    3. Cluster them once using `peptml.HDBSCANClusterer` to find the tracer
       locations (i.e. cluster *centres*) - this is *1-pass clustering*.
    4. Cluster the tracer locations from the previous step again, using another
       `peptml.HDBSCANClusterer` - this is *2-pass clustering*.
    5. Segregate the intertwined points from the tracer trajectories using
       `trajectory_separation.segregate_trajectories`.
    6. Reconnect the trajectories of tracers with gaps in their tracks (such as
       when they collide) based on their cluster size using
       `trajectory_separation.connect_trajectories`.

    For more details check out the PEPT-ML paper (doi.org/10.1063/1.5129251)
    and the documentation of the functions used.

    The parameters below are, of course, easy to change. Their default values
    represent some (heuristically) good starting parameters that should work in
    the majority of cases. Perhaps the most important ones to change would be
    `k1` (a larger value represents harsher clustering; lower it for very
    noisy data), `trajectory_cut_distance` (for segregating pre-tracked points)
    and `max_signature_difference` (for reconnecting trajectories with gaps).

    Attributes
    ----------
    lors : pept.LineData instance
        The input lines of response (LoRs). As they can be stored in various
        sources (CSV files, binary data, different scanner geometries), it is
        expected that users initialise them as `pept.LineData` (or a subclass
        thereof) *before* using this script.
    max_tracers : int
        The maximum number of tracers that may be present in the field of view
        at any time. Note that the PEPT-ML algorithm does not *need* this
        number (and can work well even when tracers leave the FoV); however, it
        is useful when setting the sample sizes, overlaps and clustering
        parameters.
    sample_size : int, optional
        The `lors.sample_size`, used when transforming LoRs into cutpoints. By
        default, it is set to 200 * `max_tracers`.
    overlap : int, optional
        The `lors.overlap`, used when transforming LoRs into cutpoints. By
        default, it is set to 150 * `max_tracers`.
    sample_size2 : int, optional
        The `centres.sample_size`, used for the second pass of clustering (i.e.
        re-clustering the tracer locations found from the cutpoints). By
        default, it is set to 30 * `max_tracers`; this small value is fine as
        the data is already very clean after the first pass of clustering.
    overlap2 : int, optional
        The `centres.overlap`, used for the second pass of clustering (i.e.
        re-clustering the tracer locations found from the cutpoints). By
        default, it is set to `sample_size2` - 1, such that the loss of time
        resolution is minimised.
    max_distance : float, default 0.15
        The maximum allowed distance between two LoRs for their cutpoint to be
        considered; used by `peptml.Cutpoints`.
    k1 : float, default 0.15
        The fitting parameter for setting the `min_cluster_size` of the
        `clusterer` used for the first pass of clustering. As the number of
        cutpoints around a tracer scales with the sample_size and (inversely)
        `max_tracers`, the following formula is used -
        `min_cluster_size` = int(k1 * cutpoints.sample_size / max_tracers)`.
        A larger value represents "harsher" clustering. Lower it for very noisy
        data.
    k2 : float, default 0.6
        The fitting parameter for setting the `min_cluster_size` of the
        `clusterer2` used for the second pass of clustering. As the number of
        points around a tracer scales with the sample_size and (inversely)
        `max_tracers`, the following formula is used -
        `min_cluster_size2` = int(k2 * centres.sample_size / max_tracers)`.
        A larger value represents "harsher" clustering.
    points_window : int, optional
        The number of points included in the sliding window used when
        segregating the intertwined points from multiple trajectories using
        `pept.tracking.trajectory_separation.segregate_trajectories`. By
        default, it is set to 10 * `max_tracers`.
    trajectory_cut_distance : float, default 10.0
        Cut trajectories that are further apart than this value and consider
        the remaining ones distinct trajectories. This depends on how close the
        points in the data set are. Used by
        `pept.tracking.trajectory_separation.segregate_trajectories`.
    max_time_difference : float, default 150
        Only try to re-connect trajectories with gaps if their ends are closer
        time-wise than `max_time_difference`. Used by
        `pept.tracking.trajectory_separation.connect_trajectories`. Check the
        consistency of the units used for time.
    max_signature_difference : float, default 100
        Re-connect trajectories with gaps if the difference in cluster size
        (i.e. tracer signature) of the closest points is smaller than this
        value. An average of the 50 closest points is used. Used by
        `pept.tracking.trajectory_separation.connect_trajectories`.

    cutpoints : pept.tracking.peptml.Cutpoints instance
        The cutpoints computed from `lors`. It is a subclass of
        `pept.PointData`, such that all its methods and attributes can be used.

    min_cluster_size : int
        The minimum cluster size parameter used by the clusterer for the first
        pass of clustering.
    clusterer : pept.tracking.peptml.HDBSCANClusterer instance
        The clusterer used for the first pass of clustering. It also appends
        the "cluster size" (i.e. the number of points in the cluster
        surrounding a tracer) to `centres`.
    centres : pept.PointData instance
        The tracer locations (i.e. cluster *centres*) found after the first
        pass of clustering. The last column represents the cluster size; that
        is, the number of points in the cluster surrounding the tracer. This
        can act as a tracer signature for when their identity is temporarily
        lost, like when they collide.
    labelled_cutpoints : pept.PointData instance
        The cutpoints that were clustered, including their cluster labels as
        the last column.

    min_cluster_size2 : int
        The minimum cluster size parameter used by the clusterer for the second
        pass of clustering.
    clusterer2 : pept.tracking.peptml.HDBSCANClusterer instance
        The clusterer used for the second pass of clustering.
    centres2 : pept.PointData instance
        The tracer locations (i.e. cluster *centres*) found after the second
        pass of clustering.
    labelled_cetres : pept.PointData instance
        The centres that were clustered, including their cluster labels as the
        last column.

    segregated_trajectories : pept.PointData instance
        The segregated trajectories found from the intertwined tracer locations
        in `centres2`. It contains the same data as `centres2`, but with an
        added column representing the trajectory label (or index); they are
        indexed from 0, with -1 signifying noise. Returned by
        `pept.tracking.trajectory_separation.segregate_trajectories`.
    trajectories : pept.PointData instance
        The final separated trajectories, after reconnecting the
        `segregated_trajectories` based on their cluster size using
        `pept.tracking.trajectory_separation.connect_trajectories`. It contains
        the same data as `segregated_trajectories`, but the last column (i.e.
        trajectory label) is changed to reflect the final trajectories.
    trajectory_list : list of numpy.ndarray
        The trajectories from `trajectories` are separated into a list of
        individual trajectories, based on their last column (i.e. trajectory
        label).

    grapher : pept.visualisation.PlotlyGrapher instance
        The grapher organising the 3D subplots.

    Methods
    -------
    typecheck_and_create_attributes(\
        lors,\
        max_tracers,\
        sample_size = None,\
        overlap = None,\
        sample_size2 = None,\
        overlap2 = None,\
        max_distance = 0.15,\
        k1 = 0.15,\
        k2 = 0.6,\
        points_window = None,\
        trajectory_cut_distance = 10.0,\
        max_time_difference = 150.0,\
        max_signature_difference = 100.0\
    )
        Typecheck input parameters and create class attributes. Used by the
        class constructor for checks that are not relevant to the actual
        PEPT-ML aglorithm.
    minimal(\
        lors,\
        max_tracers,\
        sample_size = None,\
        overlap = None,\
        sample_size2 = None,\
        overlap2 = None,\
        max_distance = 0.15,\
        k1 = 0.15,\
        k2 = 0.6,\
        points_window = None,\
        trajectory_cut_distance = 10.0,\
        max_time_difference = 150,\
        max_signature_difference = 100\
    )
        The minimal code for using the PEPT-ML algorithm, without
        pretty-printing to the terminal. Useful for copy-pasting the code and
        hacking into it.

    Raises
    ------
    TypeError
        If `lors` is not an instance of `pept.LineData` (or subclass thereof!).
    AssertionError
        If `max_tracers` is smaller than 1.

    Notes
    -----
    The `pept` package is modular, meaning it can be used in many ways, to suit
    (m)any circumstances. This is just an example, which you can modify as you
    see fit.

    If you're not sure what a class or method does, you can always consult the
    documentation online, at "uob-positron-imaging-centre.github.io", or by
    adding a questions mark in front of a package / module / class / function
    name in the Python shell or using the `help` command. For example:

    >>> import pept
    >>> ?pept.LineData          # Short documentation
    >>> ??pept.LineData         # Long documentation
    >>> help(pept.LineData)     # Even more info

    Examples
    --------
    An example call would be:

    >>> lors_raw = pept.utilities.read_csv("csv_file.csv")  # two tracers CSV
    >>> lors = pept.LineData(lors_raw)
    >>> user = pept.cookbook.PEPTMLUser(lors, 2)            # max_tracers = 2

    '''

    def __init__(
        self,
        lors,                               # Essential, pept.LineData
        max_tracers,                        # Essential, int
        sample_size = None,                     # Default 200 * max_tracers
        overlap = None,                         # Default 150 * max_tracers
        sample_size2 = None,                    # Default 30 * max_tracers
        overlap2 = None,                        # Default sample_size2 - 1
        max_distance = 0.15,                # mm
        k1 = 0.15,                          # non-dimensional, important
        k2 = 0.6,                           # non-dimensional
        points_window = None,                   # Default 10 * max_tracers
        trajectory_cut_distance = 10.0,     # mm, important
        max_time_difference = 150,          # ms
        max_signature_difference = 100      # cluster size difference
    ):
        '''PEPTMLUser class constructor. This is the PEPT-ML algorithm with
        pretty-printing of the relevant objects to the terminal.

        Parameters
        ----------
        lors : pept.LineData instance
            The input lines of response (LoRs). As they can be stored in
            various sources (CSV files, binary data, different scanner
            geometries), it is expected that users initialise them as
            `pept.LineData` (or a subclass thereof) *before* using this script.
        max_tracers : int
            The maximum number of tracers that may be present in the field of
            view at any time. Note that the PEPT-ML algorithm does not *need*
            this number (and can work well even when tracers leave the FoV);
            however, it is useful to know when setting the sample sizes,
            overlaps and clustering parameters.
        sample_size : int, optional
            The `lors.sample_size`, used when transforming LoRs into cutpoints.
            By default, it is set to 200 * `max_tracers`.
        overlap : int, optional
            The `lors.overlap`, used when transforming LoRs into cutpoints. By
            default, it is set to 150 * `max_tracers`.
        sample_size2 : int, optional
            The `centres.sample_size`, used for the second pass of clustering
            (i.e. re-clustering the tracer locations found from the cutpoints).
            By default, it is set to 30 * `max_tracers`; this small value is
            fine as the data is already very clean after the first pass of
            clustering.
        overlap2 : int, optional
            The `centres.overlap`, used for the second pass of clustering (i.e.
            re-clustering the tracer locations found from the cutpoints). By
            default, it is set to `sample_size2` - 1, such that the loss of
            time resolution is minimised.
        max_distance : float, default 0.15
            The maximum allowed distance between two LoRs for their cutpoint to
            be considered; used by `peptml.Cutpoints`.
        k1 : float, default 0.15
            The fitting parameter for setting the `min_cluster_size` of the
            `clusterer` used for the first pass of clustering. As the number of
            cutpoints around a tracer scales with the sample_size and
            (inversely) `max_tracers`, the following formula is used -
            `min_cluster_size` = int(k1 * cutpoints.sample_size / max_tracers)`
            A larger value represents "harsher" clustering. Lower it for very
            noisy data.
        k2 : float, default 0.6
            The fitting parameter for setting the `min_cluster_size` of the
            `clusterer2` used for the second pass of clustering. As the number
            of points around a tracer scales with the sample_size and
            (inversely) `max_tracers`, the following formula is used -
            `min_cluster_size2` = int(k2 * centres.sample_size / max_tracers)`.
            A larger value represents "harsher" clustering.
        points_window : int, optional
            The number of points included in the sliding window used when
            segregating the intertwined points from multiple trajectories using
            `pept.tracking.trajectory_separation.segregate_trajectories`. By
            default, it is set to 10 * `max_tracers`.
        trajectory_cut_distance : float, default 10.0
            Cut trajectories that are further apart than this value and
            consider the remaining ones as distinct trajectories. This depends
            on how close the points in the data set are. Used by
            `pept.tracking.trajectory_separation.segregate_trajectories`.
        max_time_difference : float, default 150
            Only try to re-connect trajectories with gaps if their ends are
            closer time-wise than `max_time_difference`. Used by
            `pept.tracking.trajectory_separation.connect_trajectories`. Check
            the consistency of the units used for time.
        max_signature_difference : float, default 100
            Re-connect trajectories with gaps if the difference in cluster size
            (i.e. tracer signature) of the closest points is smaller than this
            value. An average of the 50 closest points is used. Used by
            `pept.tracking.trajectory_separation.connect_trajectories`.

        Raises
        ------
        TypeError
            If `lors` is not an instance of `pept.LineData` (or subclass
            thereof!).
        AssertionError
            If `max_tracers` is smaller than 1.
        '''

        # Initial type-checking of input parameters + properties
        # initialisation.
        self.typecheck_and_create_attributes(
            lors, max_tracers, sample_size, overlap, sample_size2, overlap2,
            max_distance, k1, k2, points_window, trajectory_cut_distance,
            max_time_difference, max_signature_difference
        )

        # Start the PEPT-ML algorithm; you can copy-paste the code below and
        # hack into it! It's recommended actually!
        print(
            bcolors.HEADER + "=" * 70 +     # Use colours when printing
            "\nPEPT-ML User Start\n" +
            "=" * 70 + bcolors.ENDC         # End colour
        )

        # Print input LoRs and max_tracers. For `lors`, print its printable
        # representation (i.e. long representation) using repr().
        self.lors.sample_size = self.sample_size
        self.lors.overlap = self.overlap
        print((
            f"\n{bcolors.OKBLUE}"
            f"LoRs:{bcolors.ENDC}\n"
            f"{repr(self.lors)}\n"
        ))
        print((
            f"{bcolors.OKBLUE}"
            f"max_tracers = {self.max_tracers}"
            f"{bcolors.ENDC}\n\n"
        ))

        # Transform the LoRs into cutpoints using the `Cutpoints` class. It is
        # a `pept.PointData` subclass, so we inherit all its attributes and
        # methods!
        print(
            bcolors.HEADER + "-" * 70 +
            "\nTransforming LoRs into Cutpoints\n" +
            "-" * 70 + bcolors.ENDC
        )
        print((
            f"\n{bcolors.OKBLUE}"
            f"max_distance = {self.max_distance}"
            f"{bcolors.ENDC}\n"
        ))

        self.cutpoints = peptml.Cutpoints(lors, max_distance)
        print((
            f"\n{bcolors.OKGREEN}"
            f"Finding cutpoints completed.\n\n"
            f"{bcolors.OKBLUE}"
            f"Cutpoints:{bcolors.ENDC}\n"
            f"{repr(self.cutpoints)}\n\n"
        ))

        # Find the tracer locations using peptml.HDBSCANClusterer. This is
        # 1-pass clustering.
        print(
            bcolors.HEADER + "-" * 70 +
            "\nClustering Cutpoints using HDBSCANClusterer " +
            "(1-pass clustering)\n" +
            "-" * 70 + bcolors.ENDC
        )

        # The cluster size (i.e. number of cutpoints around each tracer
        # location) scales with the maximum number of tracers that might be
        # present in the field of view at any time.
        self.min_cluster_size = int(self.k1 * self.cutpoints.sample_size /
                                    self.max_tracers)
        self.clusterer = peptml.HDBSCANClusterer(self.min_cluster_size)
        print((
            f"\n{bcolors.OKBLUE}"
            f"Clusterer for 1-pass clustering:{bcolors.ENDC}\n"
            f"{repr(self.clusterer)}\n"
        ))

        self.centres, self.labelled_cutpoints = self.clusterer.fit(
            self.cutpoints, get_labels = True
        )

        # Set the sample size and overlap of the tracer locations from the
        # first pass of clustering. The data is already very clean from the
        # previous steps, so we'll set a small sample size and large overlap to
        # minimise smoothing, or the temporal resolution lost.
        self.centres.sample_size = self.sample_size2
        self.centres.overlap = self.overlap2

        print((
            f"\n{bcolors.OKGREEN}"
            f"1-pass clustering completed.\n\n"
            f"{bcolors.OKBLUE}"
            f"Cluster centres (i.e. tracer locations):{bcolors.ENDC}\n"
            f"{repr(self.centres)}\n\n{bcolors.OKBLUE}"
            f"Labelled cutpoints:{bcolors.ENDC}\n"
            f"{repr(self.labelled_cutpoints)}\n\n"
        ))

        # Create smoother, tighter tracer locations using
        # peptml.HDBSCANClusterer. This is 2-pass clustering.
        print(
            bcolors.HEADER + "-" * 70 +
            "\nClustering Centres using HDBSCANClusterer " +
            "(2-pass clustering)\n" +
            "-" * 70 + bcolors.ENDC
        )

        self.min_cluster_size2 = int(self.k2 * self.centres.sample_size /
                                     self.max_tracers)
        self.clusterer2 = peptml.HDBSCANClusterer(self.min_cluster_size2)
        print((
            f"\n{bcolors.OKBLUE}"
            f"Clusterer for 2-pass clustering:{bcolors.ENDC}\n"
            f"{repr(self.clusterer2)}\n"
        ))

        self.centres2, self.labelled_centres = self.clusterer2.fit(
            self.centres, get_labels = True
        )

        print((
            f"\n{bcolors.OKGREEN}"
            f"2-pass clustering completed.\n\n"
            f"{bcolors.OKBLUE}"
            f"Tracer locations after 2-pass clustering:{bcolors.ENDC}\n"
            f"{repr(self.centres2)}\n\n"
        ))

        # Segregate the intertwined points of the previous step.
        print(
            bcolors.HEADER + "-" * 70 +
            "\nSegregate the intertwined points from the 2-pass clustering\n" +
            "-" * 70 + bcolors.ENDC
        )

        self.segregated_trajectories = tsp.segregate_trajectories(
            self.centres2, self.points_window, self.trajectory_cut_distance
        )

        print((
            f"\n{bcolors.OKGREEN}"
            f"Trajectory segregation completed.\n\n"
            f"{bcolors.OKBLUE}"
            f"Segregated trajectories:{bcolors.ENDC}\n"
            f"{repr(self.segregated_trajectories)}\n"
        ))

        print((
            f"\n{bcolors.OKBLUE}"
            "Number of trajectories found after segregation: "
            f"{self.segregated_trajectories.points[:, -1].max() + 1}\n\n"
        ))

        # Connect the segregated trajectories based on their cluster size.
        print(
            bcolors.HEADER + "-" * 70 +
            "\nConnect the segregated trajectories based on their cluster " +
            "size\n" + "-" * 70 + bcolors.ENDC
        )

        self.trajectories = tsp.connect_trajectories(
            self.segregated_trajectories,
            self.max_time_difference,
            self.max_signature_difference
        )

        self.trajectory_list = pept.utilities.group_by_column(
            self.trajectories.points, -1
        )

        print((
            f"\n{bcolors.OKGREEN}"
            f"Trajectory connecting completed.\n\n"
            f"{bcolors.OKBLUE}"
            f"Trajectories:{bcolors.ENDC}\n"
            f"{repr(self.trajectories)}\n"
        ))

        print((
            f"\n{bcolors.OKBLUE}"
            "Number of trajectories found: "
            f"{self.trajectories.points[:, -1].max() + 1}\n\n"
        ))

        # Plotting time!
        print(
            bcolors.HEADER + "-" * 70 +
            "\nPlotting Results using PlotlyGrapher\n" +
            "-" * 70 + bcolors.ENDC
        )

        # Instantiate a PlotlyGrapher instance with 2x3 subplots:
        self.grapher = PlotlyGrapher(rows = 2, cols = 3, subplot_titles = [
            "LoRs, first two samples",
            "Labelled cutpoints, first 10 samples, 1-pass",
            "Tracer locations, 1-pass (coloured cluster size)",
            "Tracer locations, 2-pass (coloured cluster size)",
            "Segregated tracer locations",
            "Individual trajectories"
        ])

        # Add traces to the grapher figure. row = 1, col = 1 by default.
        # If there are less than 2 samples of LoRs, plot all LoRs.
        if len(self.lors) >= 2:
            self.grapher.add_trace(self.lors.lines_trace([0, 1]))
        else:
            self.grapher.add_lines(self.lors)

        # If there are less than 10 samples of cutpoints, plot all of them
        if len(self.labelled_cutpoints) >= 10:
            self.grapher.add_trace(self.labelled_cutpoints.points_trace(
                range(0, 10)
            ), row = 1, col = 2)
        else:
            self.grapher.add_trace(self.labelled_cutpoints.points_trace(),
                                   col = 2)
        self.grapher.add_points(self.centres, row = 1, col = 3)
        self.grapher.add_points(self.centres2, colorbar_col = 4,
                                row = 2, col = 1)
        self.grapher.add_points(self.segregated_trajectories, row = 2, col = 2)
        self.grapher.add_points(self.trajectories, row = 2, col = 3)

        self.grapher.show()

        print((
            f"\n{bcolors.OKGREEN}"
            f"Plotting completed."
            f"{bcolors.ENDC}\n\n"
        ))

        print(
            bcolors.HEADER + "=" * 70 +     # Use colorrs when printing
            "\nPEPT-ML User End\n" +
            "=" * 70 + bcolors.ENDC         # End colour
        )


    @staticmethod
    def minimal(
        lors,                               # Essential, pept.LineData
        max_tracers,                        # Essential, int
        sample_size = None,                     # Default 200 * max_tracers
        overlap = None,                         # Default 150 * max_tracers
        sample_size2 = None,                    # Default 30 * max_tracers
        overlap2 = None,                        # Default sample_size2 - 1
        max_distance = 0.15,                # mm
        k1 = 0.15,                          # non-dimensional, important
        k2 = 0.6,                           # non-dimensional
        points_window = None,                   # Default 10 * max_tracers
        trajectory_cut_distance = 10.0,     # mm, important
        max_time_difference = 150,          # ms
        max_signature_difference = 100      # cluster size difference
    ):
        # Check `lors` is an instance of `pept.LineData`; this includes any
        # subclass! (e.g. `pept.scanners.ParallelScreens`). If it's not, raise
        # an error.
        if not isinstance(lors, pept.LineData):
            raise TypeError(
                textwrap.fill((
                    "\n[ERROR]: `lors` should be an instance of `pept."
                    f"LineData` (or subclass thereof). Received {type(lors)}."
                    "Note: you can read in LoR data from a CSV file using "
                    "`pept.utilities.read_csv`; initialise them in a `pept."
                    "LineData` before calling this function. Check this "
                    "script's documentation for more.\n"
                ), replace_whitespace = False)
            )

        # If any of sample_size, overlap, sample_size2, overlap2 were left
        # to the default `None`, set them to the default values.
        if sample_size is None:
            sample_size = 200 * max_tracers
        if overlap is None:
            overlap = 150 * max_tracers
        if sample_size2 is None:
            sample_size2 = 30 * max_tracers
        if overlap2 is None:
            overlap2 = sample_size2 - 1

        # Set points_window to a default value.
        if points_window is None:
            points_window = max_tracers * 10

        # Beginning of the PEPT-ML User algorithm:

        # Compute cutpoints from LoRs
        lors.sample_size = sample_size
        lors.overlap = overlap
        cutpoints = peptml.Cutpoints(lors, max_distance)

        # 1-pass clustering
        min_cluster_size = int(k1 * cutpoints.sample_size / max_tracers)
        clusterer = peptml.HDBSCANClusterer(min_cluster_size)
        centres, labelled_cutpoints = clusterer.fit(cutpoints,
                                                    get_labels = True)

        # 2-pass clustering
        centres.sample_size = sample_size2
        centres.overlap = overlap2

        min_cluster_size2 = int(k2 * centres.sample_size / max_tracers)
        clusterer2 = peptml.HDBSCANClusterer(min_cluster_size2)
        centres2, labelled_centres = clusterer2.fit(centres,
                                                    get_labels = True)

        # Segregate the intertwined points
        segregated_trajectories = tsp.segregate_trajectories(
            centres2, points_window, trajectory_cut_distance
        )

        # Connect segregated points based on cluster size
        trajectories = tsp.connect_trajectories(
            segregated_trajectories,
            max_time_difference,
            max_signature_difference
        )

        # Create a list of individual trajectories
        # trajectory_list = pept.utilities.group_by_column(
        #     trajectories.points, -1
        # )

        # Plot everything using PlotlyGrapher
        grapher = PlotlyGrapher(rows = 2, cols = 3, subplot_titles = [
            "LoRs, first two samples",
            "Labelled cutpoints, first 10 samples, 1-pass",
            "Tracer locations, 1-pass (coloured cluster size)",
            "Tracer locations, 2-pass (coloured cluster size)",
            "Segregated tracer locations",
            "Individual trajectories"
        ])

        grapher.add_trace(lors.lines_trace([0, 1]))
        grapher.add_trace(labelled_cutpoints.points_trace(range(0, 10)),
                          col = 2)
        grapher.add_points(centres, col = 3)
        grapher.add_points(centres2, colorbar_col = 4, row = 2, col = 1)
        grapher.add_points(segregated_trajectories, row = 2, col = 2)
        grapher.add_points(trajectories, row = 2, col = 3)

        grapher.show()


    def typecheck_and_create_attributes(
        self,
        lors,
        max_tracers,
        sample_size = None,
        overlap = None,
        sample_size2 = None,
        overlap2 = None,
        max_distance = 0.15,
        k1 = 0.15,
        k2 = 0.6,
        points_window = None,
        trajectory_cut_distance = 10.0,
        max_time_difference = 150.0,
        max_signature_difference = 100.0
    ):
        # Check `lors` is an instance of `pept.LineData`; this includes any
        # subclass! (e.g. `pept.scanners.ParallelScreens`). If it's not, raise
        # an error.
        if not isinstance(lors, pept.LineData):
            raise TypeError(
                textwrap.fill((
                    "\n[ERROR]: `lors` should be an instance of `pept."
                    f"LineData` (or subclass thereof). Received {type(lors)}. "
                    "Note: you can read in LoR data from a CSV file using "
                    "`pept.utilities.read_csv`; initialise them in a `pept."
                    "LineData` before calling this function. Check this "
                    "script's documentation for more information.\n"
                ), replace_whitespace = False)
            )

        # Attach lors to our class instance so we can access it as a class
        # property. Attach the next variables used too.
        self.lors = lors

        # Make sure max_tracers is an `int` that's larger or equal to 1
        max_tracers = int(max_tracers)
        assert max_tracers >= 1, textwrap.fill(
            "[ERROR]: `max_tracers` should be an `int` >= 1."
        )
        self.max_tracers = max_tracers

        # If any of sample_size, overlap, sample_size2, overlap2 were left
        # to the default `None`, set them to the default values. Otherwise
        # type-check them; also set lors.sample_size and lors.overlap.
        if sample_size is None:
            sample_size = 200 * max_tracers
        else:
            sample_size = int(sample_size)

        self.sample_size = sample_size

        if overlap is None:
            overlap = 150 * max_tracers
        else:
            overlap = int(overlap)

        self.overlap = overlap

        if sample_size2 is None:
            sample_size2 = 30 * max_tracers
        else:
            sample_size2 = int(sample_size2)

        self.sample_size2 = sample_size2

        if overlap2 is None:
            overlap2 = sample_size2 - 1
        else:
            int(overlap2)

        self.overlap2 = overlap2

        # Type-check max_distance (for cutpoints), k1 (for 1-pass clusterer)
        # and k2 (for the 2-pass clusterer)
        max_distance = float(max_distance)
        self.max_distance = max_distance

        k1 = float(k1)
        self.k1 = k1

        k2 = float(k2)
        self.k2 = k2

        # Type-check parameters for segregate_trajectories
        if points_window is None:
            self.points_window = self.max_tracers * 10
        else:
            self.points_window = int(points_window)

        self.trajectory_cut_distance = float(trajectory_cut_distance)

        # Type-check parameters for connect_trajectories
        self.max_time_difference = float(max_time_difference)
        self.max_signature_difference = float(max_signature_difference)




class PEPTMLFindParameters:
    '''This is an example script that visually helps find the best clustering
    parameters for the PEPT-ML algorithm from the `pept` library.

    The script is encapsulated in a class such that, after running it (i.e.
    instantiating the class), all the relevant analysis objects can be accessed
    as class attributes (e.g. LoRs, clusterers, tracer locations).

    The class has three methods:

        1. __init__ : the class constructor that is run when instantiating it;
           it contains the actual PEPT-ML algorithm.
        2. typecheck_and_create_attributes : typecheck input parameters and
           create class attributes.
        3. minimal : the minimal PEPT-ML code, without pretty-printing to the
           terminal.

    It is more than recommended to look through the code, copy and hack into
    it! The class constructor (`__init__`) contains the PEPT-ML algorithm with
    printing functionality to the terminal; if you'd like the code without such
    additions, use the equivalent code in the `minimal` method.

    The main steps of the PEPT-ML-User algorithm are:

    -- Done beforehand, by the user --

        1. Initialise LoRs in a `pept.LineData` instance (if they are stored in
           a file, you can use `pept.utilities.read_csv`). **You must do this
           before using this script**.

    -- Done by this script --

        2. Compute the cutpoints from the LoRs using the `peptml.Cutpoints`
           class.
        3. For every value in `np.linspace(k1[0], k1[1], iterations)`, cluster
           the cutpoints using `peptml.HDBSCANClusterer` to find the tracer
           locations (i.e. cluster *centres*) - this is *1-pass clustering*.

    The function then plots the clustered cutpoints with colour-coded cluster
    labels and tracer locations for each value of k1 that was chosen. The user
    can then visually select the value of k1 that works best for clustering the
    given dataset.

    For more details check out the PEPT-ML paper (doi.org/10.1063/1.5129251)
    and the documentation of the functions used.

    Attributes
    ----------
    lors : pept.LineData instance
        The input lines of response (LoRs). As they can be stored in various
        sources (CSV files, binary data, different scanner geometries), it is
        expected that users initialise them as `pept.LineData` (or a subclass
        thereof) *before* using this script.
    max_tracers : int
        The maximum number of tracers that may be present in the field of view
        at any time. Note that the PEPT-ML algorithm does not *need* this
        number (and can work well even when tracers leave the FoV); however, it
        is useful when setting the sample sizes, overlaps and clustering
        parameters.
    sample_size : int, optional
        The `lors.sample_size`, used when transforming LoRs into cutpoints. By
        default, it is set to 200 * `max_tracers`.
    overlap : int, optional
        The `lors.overlap`, used when transforming LoRs into cutpoints. By
        default, it is set to 150 * `max_tracers`.
    max_distance : float, default 0.15
        The maximum allowed distance between two LoRs for their cutpoint to be
        considered; used by `peptml.Cutpoints`.
    k1 : list[float], default [0.05, 0.8]
        The range to explor of the fitting parameter for setting the
        `min_cluster_size` of the `clusterer` used for the first pass of
        clustering. As the number of cutpoints around a tracer scales with the
        sample_size and (inversely) `max_tracers`, the following formula is
        used -
        `min_cluster_size` = int(k1 * cutpoints.sample_size / max_tracers)`.
        A larger value represents "harsher" clustering. This class will iterate
        through `iterations` equally-spaced values between k1[0] and k1[1] for
        one-pass clustering.
    iterations : int, default 5
        The number of values of k1 that will be taken between k1[0] and k1[1]
        for the different clusterers.

    cutpoints : pept.tracking.peptml.Cutpoints instance
        The cutpoints computed from `lors`. It is a subclass of
        `pept.PointData`, such that all its methods and attributes can be used.

    ks : np.ndarray
        The values of k1 that will be used for the clusterers. It is equivalent
        to a call to `np.linspace(k1[0], k1[1], iterations, dtype = float)`.
    clusterers : list[pept.tracking.peptml.HDBSCANClusterer]
        The list of clusterers with different `min_cluster_size` values from
        `ks` that will be used for one-pass clustering.

    centres : list[pept.PointData]
        A list of the tracer locations (i.e. cluster *centres*) found after the
        first pass of clustering, for every value of k1 in `ks`. The last
        column in each `pept.PointData` represents the cluster size; that is,
        the number of points in the cluster surrounding the tracer. This can
        act as a tracer signature for when their identity is temporarily lost,
        like when they collide.
    labelled_cutpoints : list[pept.PointData]
        A list of the cutpoints that were clustered for every value of k1 in
        `ks`, including their cluster labels as the last column.

    grapher : pept.visualisation.PlotlyGrapher instance
        The grapher organising the 3D subplots.

    Methods
    -------
    typecheck_and_create_attributes(\
        lors,\
        max_tracers,\
        sample_size = None,\
        overlap = None,\
        max_distance = 0.15,\
        k1 = [0.05, 0.8],\
        iterations = 5\
    )
        Typecheck input parameters and create class attributes. Used by the
        class constructor for checks that are not relevant to the actual
        PEPT-ML aglorithm.
    minimal(\
        lors,\
        max_tracers,\
        sample_size = None,\
        overlap = None,\
        max_distance = 0.15,\
        k1 = [0.05, 0.8],\
        iterations = 5\
    )
        The minimal code for using the algorithm in this class, without
        pretty-printing to the terminal. Useful for copy-pasting the code and
        hacking into it.

    Raises
    ------
    TypeError
        If `lors` is not an instance of `pept.LineData` (or subclass thereof!).
    AssertionError
        If `max_tracers` is smaller than 1.

    Notes
    -----
    The `pept` package is modular, meaning it can be used in many ways, to suit
    (m)any circumstances. This is just an example, which you can modify as you
    see fit.

    If you're not sure what a class or method does, you can always consult the
    documentation online, at "uob-positron-imaging-centre.github.io", or by
    adding a questions mark in front of a package / module / class / function
    name in the Python shell or using the `help` command. For example:

    >>> import pept
    >>> ?pept.LineData          # Short documentation
    >>> ??pept.LineData         # Long documentation
    >>> help(pept.LineData)     # Even more info

    Examples
    --------
    An example call would be:

    >>> lors_raw = pept.utilities.read_csv("csv_file.csv")  # two tracers CSV
    >>> lors = pept.LineData(lors_raw)
    >>> user = pept.cookbook.PEPTMLFindParameters(lors, 2)  # max_tracers = 2

    '''

    def __init__(
        self,
        lors,                       # Essential, pept.LineData
        max_tracers,                # Essential, int
        sample_size = None,             # Default 200 * max_tracers
        overlap = None,                 # Default 150 * max_tracers
        max_distance = 0.15,        # mm
        k1 = [0.05, 0.8],           # non-dimensional, important
        iterations = 5              # number of points between k1[0] and k1[1]
    ):
        '''The PEPTMLFindParameters class constructor.

        Parameters
        ----------
        lors : pept.LineData instance
            The input lines of response (LoRs). As they can be stored in
            various sources (CSV files, binary data, different scanner
            geometries), it is expected that users initialise them as
            `pept.LineData` (or a subclass thereof) *before* using this script.
        max_tracers : int
            The maximum number of tracers that may be present in the field of
            view at any time. Note that the PEPT-ML algorithm does not *need*
            this number (and can work well even when tracers leave the FoV);
            however, it is useful when setting the sample sizes, overlaps and
            clustering parameters.
        sample_size : int, optional
            The `lors.sample_size`, used when transforming LoRs into cutpoints.
            By default, it is set to 200 * `max_tracers`.
        overlap : int, optional
            The `lors.overlap`, used when transforming LoRs into cutpoints. By
            default, it is set to 150 * `max_tracers`.
        max_distance : float, default 0.15
            The maximum allowed distance between two LoRs for their cutpoint to
            be considered; used by `peptml.Cutpoints`.
        k1 : list[float], default [0.05, 0.8]
            The range to explor of the fitting parameter for setting the
            `min_cluster_size` of the `clusterer` used for the first pass of
            clustering. As the number of cutpoints around a tracer scales with
            the sample_size and (inversely) `max_tracers`, the following
            formula is used -
            `min_cluster_size` = int(k1 * cutpoints.sample_size / max_tracers)`
            A larger value represents "harsher" clustering. This class will
            iterate through `iterations` equally-spaced values between k1[0]
            and k1[1] for one-pass clustering.
        iterations : int, default 5
            The number of values of k1 that will be taken between k1[0] and
            k1[1] for the different clusterers.

        Raises
        ------
        TypeError
            If `lors` is not an instance of `pept.LineData` (or subclass
            thereof!).
        AssertionError
            If `max_tracers` is smaller than 1.
        '''

        # Initial type-checking of input parameters + properties
        # initialisation.
        self.typecheck_and_create_attributes(
            lors, max_tracers, sample_size, overlap, max_distance, k1,
            iterations
        )

        # Start the PEPT-ML algorithm; you can copy-paste the code below and
        # hack into it! It's recommended actually!
        print(
            bcolors.HEADER + "=" * 70 +     # Use colours when printing
            "\nPEPT-ML Find Parameters Start\n" +
            "=" * 70 + bcolors.ENDC         # End colour
        )

        # Print input LoRs and max_tracers. For `lors`, print its printable
        # representation (i.e. long representation) using repr().
        self.lors.sample_size = self.sample_size
        self.lors.overlap = self.overlap
        print((
            f"\n{bcolors.OKBLUE}"
            f"LoRs:{bcolors.ENDC}\n"
            f"{repr(self.lors)}\n"
        ))
        print((
            f"{bcolors.OKBLUE}"
            f"max_tracers = {self.max_tracers}"
            f"{bcolors.ENDC}\n\n"
        ))

        # Transform the LoRs into cutpoints using the `Cutpoints` class. It is
        # a `pept.PointData` subclass, so we inherit all its attributes and
        # methods!
        print(
            bcolors.HEADER + "-" * 70 +
            "\nTransforming LoRs into Cutpoints\n" +
            "-" * 70 + bcolors.ENDC
        )
        print((
            f"\n{bcolors.OKBLUE}"
            f"max_distance = {self.max_distance}"
            f"{bcolors.ENDC}\n"
        ))

        self.cutpoints = peptml.Cutpoints(lors, max_distance)
        print((
            f"\n{bcolors.OKGREEN}"
            f"Finding cutpoints completed.\n\n"
            f"{bcolors.OKBLUE}"
            f"Cutpoints:{bcolors.ENDC}\n"
            f"{repr(self.cutpoints)}\n\n"
        ))

        # Cluster our cutpoints using peptml.HDBSCANClusterer for every value
        # of k1 between k1[0] and k1[1], at `iterations` points.
        print(
            bcolors.HEADER + "-" * 70 +
            "\nClustering Cutpoints using HDBSCANClusterer " +
            "(1-pass clustering)\n" +
            "-" * 70 + bcolors.ENDC
        )

        self.ks = np.linspace(k1[0], k1[1], iterations, dtype = float)
        print((
            "\nFormula for setting the clusterer min_cluster_size:\n"
            "    min_cluster_size = int(k1 * cutpoints.sample_size / "
            "max_tracers)\n\n"
            f"{bcolors.OKBLUE}For values of k1 from:\n"
            f"    {self.ks}{bcolors.ENDC}\n\n"
        ))

        # Create a list of clusterers.
        self.clusterers = [peptml.HDBSCANClusterer(
            int(k * self.cutpoints.sample_size / self.max_tracers)
        ) for k in self.ks]

        # For each clusterer, fit the cutpoints and append the results to a
        # list. Use `tqdm` to show a progress bar.
        self.centres = []
        self.labelled_cutpoints = []
        for clusterer in tqdm(self.clusterers):
            centres1, labelled_cutpoints1 = clusterer.fit(
                self.cutpoints,
                get_labels = True,
                verbose = False
            )

            self.centres.append(centres1)
            self.labelled_cutpoints.append(labelled_cutpoints1)

        print((
            f"\n{bcolors.OKGREEN}"
            f"Clustering cutpoints for every value of k1 completed.\n\n"
            f"{bcolors.OKBLUE}" +
            textwrap.fill(
                "Tracer locations (i.e. cluster centres) for every value of "
                "k1 were saved as a list in the `centres` attribute of this "
                "class. Similarily, the labelled cutpoints (i.e. the "
                "cutpoints with an extra column of cluster labels) were saved "
                "as a list in the `labelled_cutpoints` attribute of this "
                "class."
            ) +
            f"{bcolors.ENDC}\n\n"
        ))

        # Plotting time!
        print(
            bcolors.HEADER + "-" * 70 +
            "\nPlotting Results using PlotlyGrapher\n" +
            "-" * 70 + bcolors.ENDC
        )

        # Instantiate a PlotlyGrapher instance with 2 rows and `iterations`
        # columns:
        self.grapher = PlotlyGrapher(
            rows = 2,
            cols = self.iterations,
            subplot_titles = [
                f"k1 = {k}, labelled cutpoints" for k in self.ks
            ] + [
                f"k1 = {k}, cluster centres" for k in self.ks
            ]
        )

        # For every value of k1 we used, on the first row plot the labelled
        # cutpoints, and on the second row plot the tracer locations (i.e.
        # cluster centres).
        for i in range(self.iterations):
            self.grapher.add_trace(
                self.labelled_cutpoints[i].points_trace(), row = 1, col = i + 1
            )
            self.grapher.add_trace(
                self.centres[i].points_trace(), row = 2, col = i + 1
            )

        self.grapher.show()

        print((
            f"\n{bcolors.OKGREEN}"
            f"Plotting completed."
            f"{bcolors.ENDC}\n\n"
        ))

        print(
            bcolors.HEADER + "=" * 70 +     # Use colorrs when printing
            "\nPEPT-ML Find Parameters End\n" +
            "=" * 70 + bcolors.ENDC         # End colour
        )


    @staticmethod
    def minimal(
        lors,                       # Essential, pept.LineData
        max_tracers,                # Essential, int
        sample_size = None,             # Default 200 * max_tracers
        overlap = None,                 # Default 150 * max_tracers
        sample_size2 = None,            # Default 30 * max_tracers
        overlap2 = None,                # Default sample_size2 - 1
        max_distance = 0.15,        # mm
        k1 = [0.05, 0.8],           # non-dimensional, important
        iterations = 5              # number of points between k1[0] and k1[1]
    ):

        # Check `lors` is an instance of `pept.LineData`; this includes any
        # subclass! (e.g. `pept.scanners.ParallelScreens`). If it's not, raise
        # an error.
        if not isinstance(lors, pept.LineData):
            raise TypeError(
                textwrap.fill((
                    "\n[ERROR]: `lors` should be an instance of `pept."
                    f"LineData` (or subclass thereof). Received {type(lors)}."
                    "Note: you can read in LoR data from a CSV file using "
                    "`pept.utilities.read_csv`; initialise them in a `pept."
                    "LineData` before calling this function. Check this "
                    "script's documentation for more.\n"
                ), replace_whitespace = False)
            )

        # If any of sample_size, overlap, sample_size2, overlap2 were left
        # to the default `None`, set them to the default values.
        if sample_size is None:
            sample_size = 200 * max_tracers
        if overlap is None:
            overlap = 150 * max_tracers
        if sample_size2 is None:
            sample_size2 = 30 * max_tracers
        if overlap2 is None:
            overlap2 = sample_size2 - 1

        lors.sample_size = sample_size
        lors.overlap = overlap

        cutpoints = peptml.Cutpoints(lors, max_distance)

        ks = np.linspace(k1[0], k1[1], iterations, dtype = float)
        clusterers = [peptml.HDBSCANClusterer(
            int(k * cutpoints.sample_size / max_tracers)
        ) for k in ks]

        centres = []
        labelled_cutpoints = []
        for clusterer in tqdm(clusterers):
            centres1, labelled_cutpoints1 = clusterer.fit(
                cutpoints,
                get_labels = True,
                verbose = False
            )

            centres.append(centres1)
            labelled_cutpoints.append(labelled_cutpoints1)

        grapher = PlotlyGrapher(
            rows = 2,
            cols = iterations,
            subplot_titles = [
                f"k1 = {k}, labelled cutpoints" for k in ks
            ] + [
                f"k1 = {k}, cluster centres" for k in ks
            ]
        )

        for i in range(iterations):
            grapher.add_trace(
                labelled_cutpoints[i].points_trace(), row = 1, col = i + 1
            )
            grapher.add_trace(
                centres[i].points_trace(), row = 2, col = i + 1
            )

        grapher.show()


    def typecheck_and_create_attributes(
        self,
        lors,                       # Essential, pept.LineData
        max_tracers,                # Essential, int
        sample_size = None,             # Default 200 * max_tracers
        overlap = None,                 # Default 150 * max_tracers
        max_distance = 0.15,        # mm
        k1 = [0.05, 0.8],           # non-dimensional, important
        iterations = 5              # number of points between k1[0] and k1[1]
    ):
        # Check `lors` is an instance of `pept.LineData`; this includes any
        # subclass! (e.g. `pept.scanners.ParallelScreens`). If it's not, raise
        # an error.
        if not isinstance(lors, pept.LineData):
            raise TypeError(
                textwrap.fill((
                    "\n[ERROR]: `lors` should be an instance of `pept."
                    f"LineData` (or subclass thereof). Received {type(lors)}."
                    "Note: you can read in LoR data from a CSV file using "
                    "`pept.utilities.read_csv`; initialise them in a `pept."
                    "LineData` before calling this function. Check this "
                    "script's documentation for more.\n"
                ), replace_whitespace = False)
            )

        # Attach lors to our class instance so we can access it as a class
        # property. Attach the next variables used too.
        self.lors = lors

        # Make sure max_tracers is an `int` that's larger or equal to 1
        max_tracers = int(max_tracers)
        assert max_tracers >= 1, textwrap.fill(
            "[ERROR]: `max_tracers` should be an `int` >= 1."
        )
        self.max_tracers = max_tracers

        # If any of sample_size, overlap, sample_size2, overlap2 were left
        # to the default `None`, set them to the default values. Otherwise
        # type-check them; also set lors.sample_size and lors.overlap.
        if sample_size is None:
            sample_size = 200 * max_tracers
        self.sample_size = sample_size

        if overlap is None:
            overlap = 150 * max_tracers
        self.overlap = overlap

        # Type-check max_distance (for cutpoints), k1 (for 1-pass clusterer)
        # and k2 (for the 2-pass clusterer)
        self.max_distance = float(max_distance)

        # We will use `iterations` points between k1[0] and k1[1]. Make sure
        # k1 is list-like with length 2.
        k1 = list(k1)
        assert len(k1) == 2, textwrap.fill(
            "[ERROR]: k1 should be list-like with exactly two elements."
        )
        self.k1 = k1

        self.iterations = int(iterations)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


# File   : __init__.py
# License: License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 21.08.2019


from    .peptsim    import  Noise
from    .peptsim    import  Shape
from    .peptsim    import  Simulator


__all__ = [
    'Noise',
    'Shape',
    'Simulator'
]


__author__ =        "Andrei Leonard Nicusan"
__credits__ =       ["Andrei Leonard Nicusan", "Kit Windows-Yule", "Sam Manger"]
__license__ =       "GNU v3.0"
__version__ =       "0.1"
__maintainer__ =    "Andrei Leonard Nicusan"
__email__ =         "a.l.nicusan@bham.ac.uk"
__status__ =        "Development"








#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : __init__.py
# License:  GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 22.08.2019


'''Separate the intertwined points from pre-tracked tracer locations into
individual trajectories.

Extended Summary
----------------
A typical PEPT workflow would involve transforming LoRs into points using some
tracking algorithm. These points include all tracers moving through the system,
being intertwined (e.g. for two tracers A and B, the `point_data` array might
have two entries for A, followed by three entries for B, then one entry for A,
etc.). The points can be segregated based on distance alone using the
`segregate_trajectories` function; for well-defined trajectories of tracers
that do not collide, this may be enough to retrieve individual trajectories.
However, for tracers that do come into contact, the identity of the least
active one is usually lost; in such cases, the `connect_trajectories` function
can be used to piece back the trajectories of tracers with gaps in their tracks
using some *tracer signature* (e.g. cluster size in PEPT-ML).

Functions Provided
------------------

::

    pept.tracking.trajectory_separation
    ├── segregate_trajectories : Segregate intertwined points by distance.
    ├── connect_trajectories : Connect segregated paths by tracer signatures.
    └── trajectory_errors : Calculate deviations of tracked from real paths.

Examples
--------
Take for example two tracers that go downwards (below, 'x' is the position, and
in parantheses is the array index at which that point is found in the data
array).

::

                    Some tracking algorithm
                    -----------------------
                               |
                               V

    `points`, numpy.ndarray, shape (11, 4), columns [time, x, y, z]:
         x (0)
        x (1)                       x (2)
         x (3)                     x (4)
           x (5)                 x (7)
           x (6)                x (9)
          x (8)                 x (10)

                  Use `segregate_trajectories`
                  ----------------------------
                               |
                               V

>>> import pept.tracking.trajectory_separation as tsp
>>> points_window = 10
>>> trajectory_cut_distance = 15    # mm
>>> segregated_trajectories = tsp.segregate_trajectories(
>>>     points, points_window, trajectory_cut_distance
>>> )

::

                Labelled, segregated trajectories
                ---------------------------------
                               |
                               V

    `segregated_trajectories`, numpy.ndarray, shape (11, 5),
    columns [time, x, y, z, trajectory_label]:
         x (0, label = 0)
        x (1, label = 0)            x (2, label = 1)
         x (3, label = 0)          x (4, label = 1)
           x (5, label = 0)      x (7, label = 1)
           x (6, label = 0)     x (9, label = 1)
          x (8, label = 0)      x (10, label = 1)

The usage of `connect_trajectories` is better explained using visual aids.
Check out the tutorials online at the Birmingham Positron Imaging Centre's
GitHub repository - https://github.com/uob-positron-imaging-centre.
'''


from    .trajectory_separation      import  segregate_trajectories
from    .trajectory_separation      import  connect_trajectories
from    .trajectory_separation      import  trajectory_errors


__all__ = [
    "segregate_trajectories",
    "connect_trajectories",
    "trajectory_errors"
]


__license__ = "GNU v3.0"
__maintainer__ = "Andrei Leonard Nicusan"
__email__ = "a.l.nicusan@bham.ac.uk"
__status__ = "Beta"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : __init__.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 14.01.2020


from    .find_cutpoints             import  find_cutpoints
from    .find_minpoints             import  find_minpoints


__all__ = [
    "find_cutpoints",
    "find_minpoints",
]


__license__ = "GNU v3.0"
__maintainer__ = "Andrei Leonard Nicusan"
__email__ = "a.l.nicusan@bham.ac.uk"
__status__ = "Beta"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


# File   : __init__.py
# License: License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 22.08.2019


'''PEPT-oriented visualisation tools.

Summary
-------
Visualisation functions and classes for PEPT data, transparently working with
both `pept` base classes and raw NumPy arrays (e.g. `PlotlyGrapher.add_lines`
handles both `pept.LineData` and (N, 7) NumPy arrays).

The `PlotlyGrapher` class creates interactive, publication-ready 3D figures
with optional subplots which can also be exported to portable HTML files. The
`PlotlyGrapher2D` class is its two-dimensional counterpart, handling e.g.
`pept.Pixels`.

Modules Provided
----------------

::

    pept.visualisation
    │
    Classes imported into the subpackage root:
    ├── PlotlyGrapher :     Plotly-based interactive 3D graphs with subplots.
    ├── PlotlyGrapher2D :   Plotly-based interactive 2D graphs with subplots.
    │
    Modules provided:
    └── plotly_grapher

'''


from    .plotly_grapher     import  PlotlyGrapher
from    .plotly_grapher2d   import  PlotlyGrapher2D


__all__ = [
    "PlotlyGrapher",
    "PlotlyGrapher2D",
]


__license__ = "GNU v3.0"
__maintainer__ = "Andrei Leonard Nicusan"
__email__ = "a.l.nicusan@bham.ac.uk"
__status__ = "Beta"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : __version__.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 23.08.2019


VERSION = (0, 3, 1)

__version__ = '.'.join(map(str, VERSION))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : __init__.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 10.06.2020


from    .aggregate  import  group_by_column
from    .read_csv   import  number_of_lines
from    .read_csv   import  read_csv
from    .read_csv   import  read_csv_chunks
from    .read_csv   import  ChunkReader


__all__ = [
    "group_by_column",
    "number_of_lines",
    "read_csv",
    "read_csv_chunks",
    "ChunkReader"
]


__license__ = "GNU v3.0"
__maintainer__ = "Andrei Leonard Nicusan"
__email__ = "a.l.nicusan@bham.ac.uk"
__status__ = "Beta"



# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, we ask you to cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2021 the pept developers.
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools


# File              : binary_converter.pyx
# License           : GNU v3.0
# Author            : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date              : 01.04.2021


# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: initializedcheck=False
# cython: nonecheck=False
# cython: embedsignature=True
# cython: cdivision=True


import numpy as np      # import numpy for Python functions
cimport numpy as np     # import numpy for C functions (numpy's C API)


np.import_array()


cdef extern from "binary_converter_ext.c":
    # C is included here so that it doesn't need to be compiled externally
    pass


cdef extern from "binary_converter_ext.h":
    double* read_adac_binary(const char *, Py_ssize_t *) nogil


cpdef convert_adac_forte(filepath):
    '''Convert an ADAC Forte list mode binary file to a general line data
    format `[time, x1, y1, z1, x2, y2, z2]`, returned as a NumPy array.

    ::

        Function signature:
            binary_converter(filepath)

    Binary converter for the ADAC Forte dual-head gamma camera native list
    mode data. Given the `filepath` to such a binary file (usually with
    extension ".da01"), this function converts the binary contents to the
    general line of response format `[time, x1, y1, z1, x2, y2, z2]`, where
    `z1 = 0` and `z2 = screen_separation` (found from the file).

    The LoRs are returned as a (N, 7) NumPy array, where N is the number of
    LoRs that were found in the file.

    Function parameters
    -------------------
    filepath: str-like
        A string of characters containing the path to the binary file. The
        string's contents are not read in - it will only be used with the
        `fopen` function, so it can contain any characters allowed by the OS
        file system.

    Returns
    -------
    lors: (N, 7) NumPy array
        The 2D array of LoRs, each row containing the time and coordinates of
        the first and second point defining a 3D line, respectively:
        `[time, x1, y1, z1, x2, y2, z2]`.

    Examples
    --------

    >>> import numpy as np
    >>> from pept.scanners.parallel_screens import binary_converter
    >>>
    >>> lines = binary_converter("adac_experiment_data.da01")

    '''

    filepath_utf = str(filepath).encode('UTF-8')

    cdef char                       *filepath_c = filepath_utf

    cdef double                     *lors = NULL
    cdef Py_ssize_t                 lors_elements = 0
    cdef np.npy_intp[2]             shape

    cdef np.ndarray[double, ndim=2] lors_arr

    with nogil:
        lors = read_adac_binary(filepath_c, &lors_elements)

        shape[0] = lors_elements // 7
        shape[1] = 7

    # Use the `lors` pointer as the internal data of a numpy array with
    # PyArray_SimpleNewFromData
    cdef extern from "numpy/arrayobject.h":
        void PyArray_ENABLEFLAGS(np.ndarray arr, int flags)

    if lors is NULL:
        lors_arr = np.PyArray_Empty(2, shape, np.NPY_FLOAT64, 0)
    else:
        lors_arr = np.PyArray_SimpleNewFromData(2, shape, np.NPY_FLOAT64, lors)

    PyArray_ENABLEFLAGS(lors_arr, np.NPY_OWNDATA)

    return lors_arr



# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you must cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools


# File              : traverse2d.pyx
# License           : GNU v3.0
# Author            : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date              : 03.02.2019


# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: initializedcheck=False
# cython: nonecheck=False
# cython: embedsignature=True
# cython: cdivision=True


from libc.float cimport DBL_MAX


cdef inline double fabs(double x) nogil:
    return (x if x >= 0 else -x)


cdef inline void swap(double *x, double *y) nogil:
    cdef double aux

    aux = x[0]
    x[0] = y[0]
    y[0] = aux


cdef double intersect(
    double[2] u,
    double[2] v,
    double xmin,
    double xmax,
    double ymin,
    double ymax,
) nogil:
    '''Given a 2D line defined as L(t) = U + t V and an axis-aligned bounding
    box (AABB), find the `t` for which the line intersects the box, if it
    exists.

    It is assumed that the line starts *outside* the AABB, so that t == 0.0 is
    only reserved for when the line does not intersect the box.
    '''

    cdef double tmin, tmax, tymin, tymax

    tmin = (xmin - u[0]) / v[0]     # Relies on IEEE FP behaviour for div by 0
    tmax = (xmax - u[0]) / v[0]

    tymin = (ymin - u[1]) / v[1]
    tymax = (ymax - u[1]) / v[1]

    if tmin > tmax: swap(&tmin, &tmax)
    if tymin > tymax: swap(&tymin, &tymax)

    if tmin > tymax or tymin > tmax: return 0.0

    if tymin > tmin: tmin = tymin
    if tymax < tmax: tmax = tymax

    return tmin


cpdef void traverse2d(
    double[:, :] pixels,        # Initialised to zero!
    double[:, :] lines,         # Has exactly 5 columns!
    double[:] grid_x,           # Has pixels.shape[0] + 1 elements!
    double[:] grid_y,           # Has pixels.shape[1] + 1 elements!
) nogil:
    ''' Fast pixel traversal for 2D lines (or LoRs).

    ::

        Function Signature:
            traverse2d(
                double[:, :] pixels,        # Initialised to zero!
                double[:, :] lines,         # Has exactly 7 columns!
                double[:] grid_x,           # Has pixels.shape[0] + 1 elements!
                double[:] grid_y,           # Has pixels.shape[1] + 1 elements!
            )

    This function computes the number of lines that passes through each pixel,
    saving the result in `pixels`. It does so in an efficient manner, in which
    for every line, only the pixels that it passes through are traversed.

    As it is highly optimised, this function does not perform any checks on the
    validity of the input data. Please check the parameters before calling
    `traverse2d`, as it WILL segfault on wrong input data. Details are given
    below, along with an example call.

    Parameters
    ----------
    pixels : numpy.ndarray(dtype = numpy.float64, ndim = 2)
        The `pixels` parameter is a numpy.ndarray of shape (X, Y) that has been
        initialised to zeros before the function call. The values will be
        modified in-place in the function to reflect the number of lines that
        pass through each pixel.

    lines : numpy.ndarray(dtype = numpy.float64, ndim = 2)
        The `lines` parameter is a numpy.ndarray of shape(N, 5), where each row
        is formatted as [time, x1, y1, x2, y2]. Only indices 1:5 will be used
        as the two points P1 = [x1, y1] and P2 = [x2, y2] defining the line (or
        LoR).

    grid_x : numpy.ndarray(dtype = numpy.float64, ndim = 1)
        The grid_x parameter is a one-dimensional grid that delimits the pixels
        in the x-dimension. It must be *sorted* in ascending order with
        *equally-spaced* numbers and length X + 1 (pixels.shape[0] + 1).

    grid_y : numpy.ndarray(dtype = numpy.float64, ndim = 1)
        The grid_y parameter is a one-dimensional grid that delimits the pixels
        in the y-dimension. It must be *sorted* in ascending order with
        *equally-spaced* numbers and length Y + 1 (pixels.shape[1] + 1).

    Examples
    --------
    The input parameters can be easily generated using numpy before calling the
    function. For example, if a plane of 300 x 400 is split into
    30 x 40 pixels, a possible code would be:

    >>> import numpy as np
    >>> from pept.utilities.traverse import traverse2d
    >>>
    >>> plane = [300, 400]
    >>> number_of_pixels = [30, 40]
    >>> pixels = np.zeros(number_of_pixels)

    The grid has one extra element than the number of pixels. For example, 5
    pixels between 0 and 5 would be delimited by the grid [0, 1, 2, 3, 4, 5]
    which has 6 elements (see off-by-one errors - story of my life).

    >>> grid_x = np.linspace(0, plane[0], number_of_pixels[0] + 1)
    >>> grid_y = np.linspace(0, plane[1], number_of_pixels[1] + 1)
    >>>
    >>> random_lines = np.random.random((100, 5)) * 100

    Calling `traverse2d` will modify `pixels` in-place.

    >>> traverse2d(pixels, random_lines, grid_x, grid_y)

    Notes
    -----
    This function is an adaptation of a widely-used algorithm [1]_, optimised
    for PEPT LoRs traversal.

    .. [1] Amanatides J, Woo A. A fast voxel traversal algorithm for ray tracing.
       InEurographics 1987 Aug 24 (Vol. 87, No. 3, pp. 3-10).

    '''

    n_lines = lines.shape[0]
    cdef Py_ssize_t nx = pixels.shape[0]
    cdef Py_ssize_t ny = pixels.shape[1]

    # Grid size
    cdef double gsize_x = grid_x[1] - grid_x[0]
    cdef double gsize_y = grid_y[1] - grid_y[0]

    # Delimiting grid
    cdef double xmin = grid_x[0]
    cdef double xmax = grid_x[nx]

    cdef double ymin = grid_y[0]
    cdef double ymax = grid_y[ny]

    # The current pixel indices [ix, iy] that the line passes
    # through.
    cdef Py_ssize_t ix, iy

    # Define a line as L(t) = U + t V
    # If an LoR is defined as two points P1 and P2, then
    # U = P1 and V = P2 - P1
    cdef double[2] p1, p2, u, v

    # The step [step_x, step_y, step_z] defines the sense of the LoR.
    # If V[0] is positive, then step_x = 1
    # If V[0] is negative, then step_x = -1
    cdef Py_ssize_t step_x, step_y

    # The value of t at which the line passes through to the next
    # pixel, for each dimension.
    cdef double tnext_x, tnext_y

    # deltat indicates how far along the ray we must move (in units of
    # t) for each component to be equal to the size of the pixel in
    # that dimension.
    cdef double deltat_x, deltat_y

    cdef Py_ssize_t i, j

    for i in range(n_lines):

        for j in range(2):
            p1[j] = lines[i, 1 + j]
            p2[j] = lines[i, 3 + j]

            u[j] = p1[j]
            v[j] = p2[j] - p1[j]

        ##############################################################
        # Initialisation stage

        step_x = 1 if v[0] >= 0 else -1
        step_y = 1 if v[1] >= 0 else -1

        # If the first point is outside the box, find the first pixel it hits
        if (u[0] < xmin or u[0] > xmax or
                u[1] < ymin or u[1] > ymax):

            t = intersect(u, v, xmin, xmax, ymin, ymax)

            # No intersection
            if t == 0.0:
                continue

            # Overwrite U to correspond to the first intersection point
            for j in range(2):
                u[j] = u[j] + t * v[j]

        # Corner case: every pixel is defined as lower boundary (inclusive) and
        # upper boundary (exclusive). Therefore, at the upper end of the pixel
        # grid an undefined case occurs. If a point lies right at the upper
        # boundary of the pixel space, "move it" a bit lower on the line
        if u[0] == xmax or u[1] == ymax:
            for j in range(2):
                u[j] = u[j] + 1e-5 * v[j]

        # If, for dimension x, there are 5 pixels between coordinates 0
        # and 5, then the delimiting grid is [0, 1, 2, 3, 4, 5].
        # If the line starts at 1.5, then it is part of the pixel at
        # index 1

        ix = <int>(u[0] / gsize_x)
        iy = <int>(u[1] / gsize_y)

        # Check the indices are inside the pixel grid
        if (ix < 0 or ix >= nx or iy < 0 or iy >= ny):
            continue

        # If the line is going "up", the next pixel is the next one
        # If the line is going "down", the next pixel is the current one
        if v[0] > 0:
            tnext_x = (grid_x[ix + 1] - u[0]) / v[0]
        elif v[0] < 0:
            tnext_x = (grid_x[ix] - u[0]) / v[0]
        else:
            tnext_x = DBL_MAX

        if v[1] > 0:
            tnext_y = (grid_y[iy + 1] - u[1]) / v[1]
        elif v[1] < 0:
            tnext_y = (grid_y[iy] - u[1]) / v[1]
        else:
            tnext_y = DBL_MAX

        deltat_x = fabs((grid_x[1] - grid_x[0]) / v[0]) if v[0] else 0
        deltat_y = fabs((grid_y[1] - grid_y[0]) / v[1]) if v[1] else 0

        ###############################################################
        # Incremental traversal stage

        # Loop until we reach the last pixel in space
        while (ix < nx and iy < ny) and (ix >= 0 and iy >= 0):

            pixels[ix, iy] += 1

            # Select the minimum t that makes the line pass
            # through to the next pixel
            if tnext_x < tnext_y:
                # If the next pixel falls beyond the end of the line (that is
                # at t = 1), then stop the traversal stage
                if tnext_x > 1.:
                    break

                ix = ix + step_x
                tnext_x = tnext_x + deltat_x
            else:
                if tnext_y > 1.:
                    break

                iy = iy + step_y
                tnext_y = tnext_y + deltat_y




#!/usr/bin/env python3
# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you should cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.


# File   : occupancy.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 23.11.2020




import  time
import  textwrap

import  numpy           as      np
import  pept

from    .circles_ext    import  circles2d_ext
from    .occupancy_ext  import  occupancy2d_ext


def circles2d(
    positions,
    number_of_pixels,
    radii = 0.,
    xlim = None,
    ylim = None,
    verbose = True,
):
    '''Compute the 2D occupancy of circles of different radii.

    This corresponds to the pixellisation of circular particles, such that
    each pixel's value corresponds to the area covered by the particle.

    You must specify the particles' `positions` (2D numpy array) and the
    `number_of_pixels` in each dimension (`[nx, ny, nz]`). The `radii` can be
    either:

    1. Zero: the particles are considered to be points. Each pixel will have a
       value +1 for every particle.
    2. Single positive value: all particles have the same radius.
    3. List of values of same length as `positions`: specify each particle's
       radius.

    The pixel area's bounds can be specified in `xlim` and `ylim`. If unset,
    they will be computed automatically based on the minimum and maximum
    values found in `positions`.

    Parameters
    ----------
    positions: (P, 2) numpy.ndarray
        The particles' 2D positions, where each row is formatted as
        `[x_coordinate, y_coordinate]`.

    number_of_pixels: (2,) list-like
        The number of pixels in the x-dimension and y-dimension. Each dimension
        must have at least 2 pixels.

    radii: float or (P,) list-like
        The radius of each particle. If zero, every particle is considered as
        a discrete point. If a single `float`, all particles are considered to
        have the same radius. If it is a numpy array, it specifies each
        particle's radius, and must have the same length as `positions`.

    xlim: (2,) list-like, optional
        The limits of the system over which the pixels span in the
        x-dimension, formatted as [xmin, xmax]. If unset, they will be computed
        automatically based on the minimum and maximum values found in
        `positions`.

    ylim: (2,) list-like, optional
        The limits of the system over which the pixels span in the
        y-dimension, formatted as [ymin, ymax]. If unset, they will be computed
        automatically based on the minimum and maximum values found in
        `positions`.

    verbose: bool, default True
        Time the pixellisation step and print it to the terminal.

    Returns
    -------
    pept.Pixels (numpy.ndarray subclass)
        The created pixels, each cell containing the area covered by particles.
        The `pept.Pixels` class inherits all properties and methods from
        `numpy.ndarray`, so you can use it exactly like you would a numpy
        array. It just contains extra attributes (e.g. `xlim`, `ylim`) and
        some PEPT-oriented methods (e.g. `pixels_trace`).

    Raises
    ------
    ValueError
        If `positions` is not a 2D array-like with exactly 2 columns, or
        `number_of_pixels` is not a 1D list-like with exactly 2 values or it
        contains a value smaller than 2. If `radii` is a list-like that does
        not have the same length as `positions`.

    Examples
    --------
    Create ten random particle positions between 0-100 and radii between
    0.5-2.5:

    >>> positions = np.random.random((10, 2)) * 100
    >>> radii = 0.5 + np.random.random(len(positions)) * 2

    Now pixellise those particles as circles over a grid of (20, 10) pixels:

    >>> import pept.processing as pp
    >>> num_pixels = (20, 10)
    >>> pixels = pp.circles2d(positions, num_pixels, radii)

    Alternatively, specify the system's bounds explicitly:

    >>> pixels = pp.circles2d(
    >>>     positions, (20, 10), radii, xlim = [10, 90], ylim = [-5, 105]
    >>> )

    You can plot those pixels in two ways - using `PlotlyGrapher` (this plots a
    3D "heatmap", as a coloured surface):

    >>> from pept.visualisation import PlotlyGrapher
    >>> grapher = PlotlyGrapher()
    >>> grapher.add_pixels(pixels)
    >>> grapher.show()

    Or using raw `Plotly` (this plots a "true" heatmap) - this is recommended:

    >>> import plotly.graph_objs as go
    >>> fig = go.Figure()
    >>> fig.add_trace(pixels.heatmap_trace())
    >>> fig.show()

    '''
    if verbose:
        start = time.time()

    # Type-checking inputs
    positions = np.asarray(positions, order = 'C', dtype = np.float64)

    # Check that points has at least 4 columns.
    if positions.ndim != 2 or positions.shape[1] != 2:
        raise ValueError(textwrap.fill((
            "The input `positions` should have exactly 2 columns, "
            "corresponding to the [x, y] coordinates of the particle "
            f"positions. Received array with shape {positions.shape}."
        )))

    number_of_pixels = np.asarray(
        number_of_pixels,
        order = "C",
        dtype = int
    )

    if number_of_pixels.ndim != 1 or len(number_of_pixels) != 2:
        raise ValueError(textwrap.fill((
            "The input `number_of_pixels` must be a list-like "
            "with exactly two values, corresponding to the "
            "number of pixels in the x- and y-dimension. "
            f"Received parameter with shape {number_of_pixels.shape}."
        )))

    if (number_of_pixels < 2).any():
        raise ValueError(textwrap.fill((
            "The input `number_of_pixels` must set at least two "
            "pixels in each dimension (i.e. all elements in "
            "`number_of_elements` must be larger or equal to two). "
            f"Received `{number_of_pixels}`."
        )))

    # `radii` is either a float, or a numpy array
    try:
        radii = float(radii)
        radii = np.ones(len(positions)) * radii
    except Exception:
        radii = np.asarray(radii, order = 'C', dtype = np.float64)
        if radii.ndim != 1 or len(radii) != len(positions):
            raise ValueError(textwrap.fill((
                "The input `radii` must be a float (i.e. single radius "
                "for all particles) or a 1D numpy array of radii for each "
                "particle in the system - so `radii` must have the same "
                "length as `positions`. Received array with length "
                f"{len(radii)} != {len(positions)}."
            )))

    # Maximum particle radius
    max_radius = radii.max()

    if xlim is None:
        xmin = (positions[:, 0].min() - max_radius)
        xmax = (positions[:, 0].max() + max_radius)

        xlim = np.array([
            xmin - 0.01 * abs(xmin),
            xmax + 0.01 * abs(xmax),
        ])
    else:
        xlim = np.asarray(xlim, dtype = np.float64)

        if xlim.ndim != 1 or len(xlim) != 2 or xlim[0] >= xlim[1]:
            raise ValueError(textwrap.fill((
                "The input `xlim` parameter must be a list with exactly "
                "two values, corresponding to the minimum and maximum "
                "coordinates of the pixel space in the x-dimension. "
                f"Received parameter with shape {xlim.shape}."
            )))

    if ylim is None:
        ymin = (positions[:, 1].min() - max_radius)
        ymax = (positions[:, 1].max() + max_radius)

        ylim = np.array([
            ymin - 0.01 * abs(ymin),
            ymax + 0.01 * abs(ymax),
        ])
    else:
        ylim = np.asarray(ylim, dtype = np.float64)

        if ylim.ndim != 1 or len(ylim) != 2 or ylim[0] >= ylim[1]:
            raise ValueError(textwrap.fill((
                "The input `ylim` parameter must be a list with exactly "
                "two values, corresponding to the minimum and maximum "
                "coordinates of the pixel space in the y-dimension. "
                f"Received parameter with shape {ylim.shape}."
            )))

    pixels = np.zeros(tuple(number_of_pixels), order = "C", dtype = np.float64)

    # This modifies `pixels` in-place
    circles2d_ext(
        pixels,
        positions,
        radii,
        xlim,
        ylim,
    )

    occupancy_grid = pept.Pixels(pixels, xlim = xlim, ylim = ylim)

    if verbose:
        end = time.time()
        print(f"Computed occupancy grid in {end - start} s")

    return occupancy_grid


def occupancy2d(
    points,
    number_of_pixels,
    radius,
    xlim = None,
    ylim = None,
    omit_last = False,
    verbose = True,
):
    '''Compute the 2D occupancy / residence time distribution of a single
    circular particle moving along a trajectory.

    This corresponds to the pixellisation of moving circular particles, such
    that for every two consecutive particle locations, a 2D cylinder (i.e.
    convex hull of two circles at the two particle positions), the fraction of
    its area that intersets a pixel is multiplied with the time between the
    two particle locations and saved in the input `pixels`.

    You must specify the `points` (2D numpy array) recorded along a particle's
    trajectory, formatted as [time, x, y] of each location, along with the
    `number_of_pixels` in each dimension (`[nx, ny]`) and particle `radius`.

    The pixel area's bounds can be specified in `xlim` and `ylim`. If unset,
    they will be computed automatically based on the minimum and maximum
    values found in `points`.

    Parameters
    ----------
    points: (P, 3) numpy.ndarray
        The particles' 2D locations and corresponding timestamp, where each row
        is formatted as `[time, x_coordinate, y_coordinate]`. Must have at
        least two points.

    number_of_pixels: (2,) list-like
        The number of pixels in the x-dimension and y-dimension. Each dimension
        must have at least 2 pixels.

    radius: float
        The radius of the particle. It can be given in any system of units, as
        long as it is consistent with what is used for the particle locations.

    xlim: (2,) list-like, optional
        The limits of the system over which the pixels span in the
        x-dimension, formatted as [xmin, xmax]. If unset, they will be computed
        automatically based on the minimum and maximum values found in
        `positions`.

    ylim: (2,) list-like, optional
        The limits of the system over which the pixels span in the
        y-dimension, formatted as [ymin, ymax]. If unset, they will be computed
        automatically based on the minimum and maximum values found in
        `positions`.

    omit_last: bool, default False
        If true, omit the last circle in the particle positions. Useful if
        rasterizing the same trajectory piece-wise; if you split the trajectory
        and call this function multiple times, set `omit_last = 0` to avoid
        considering the last particle location twice.

    verbose: bool, default True
        Time the pixellisation step and print it to the terminal.

    Returns
    -------
    pept.Pixels (numpy.ndarray subclass)
        The created pixels, each cell containing the area covered by particles.
        The `pept.Pixels` class inherits all properties and methods from
        `numpy.ndarray`, so you can use it exactly like you would a numpy
        array. It just contains extra attributes (e.g. `xlim`, `ylim`) and
        some PEPT-oriented methods (e.g. `pixels_trace`).

    Raises
    ------
    ValueError
        If `positions` is not a 2D array-like with exactly 3 columns, or
        `number_of_pixels` is not a 1D list-like with exactly 2 values or it
        contains a value smaller than 2. If `xlim` or `ylim` have `max` < `min`
        or there are particle positions falling outside the system defined by
        `xlim` and `ylim`, including the area.

    Examples
    --------
    Create ten random particle positions between 0-100 and radius 0.2:

    >>> positions = np.random.random((10, 2)) * 100
    >>> radius = 0.2

    Now pixellise this trajectory over a grid of (20, 10) pixels:

    >>> import pept.processing as pp
    >>> num_pixels = (20, 10)
    >>> pixels = pp.occupancy2d(positions, num_pixels, radius)

    Alternatively, specify the system's bounds explicitly:

    >>> pixels = pp.occupancy2d(
    >>>     positions, (20, 10), radius, xlim = [10, 90], ylim = [-5, 105]
    >>> )

    You can plot those pixels in two ways - using `PlotlyGrapher` (this plots a
    3D "heatmap", as a coloured surface):

    >>> from pept.visualisation import PlotlyGrapher
    >>> grapher = PlotlyGrapher()
    >>> grapher.add_pixels(pixels)
    >>> grapher.show()

    Or using raw `Plotly` (this plots a "true" heatmap) - this is recommended:

    >>> import plotly.graph_objs as go
    >>> fig = go.Figure()
    >>> fig.add_trace(pixels.heatmap_trace())
    >>> fig.show()

    '''
    if verbose:
        start = time.time()

    # Type-checking inputs
    points = np.array(points, order = 'C', dtype = np.float32)

    # Check that points has exactly 3 columns for [time, x, y]
    if points.ndim != 2 or points.shape[1] != 3:
        raise ValueError(textwrap.fill((
            "The input `points` should have exactly 3 columns, "
            "formatted as the [time, x, y] coordinates of the particle "
            f"positions. Received array with shape {points.shape}."
        )))

    times = np.array(points[:, 0], order = "C", dtype = np.float32)
    positions = np.array(points[:, 1:], order = "C", dtype = np.float32)

    number_of_pixels = np.asarray(
        number_of_pixels,
        order = "C",
        dtype = np.int32,
    )

    if number_of_pixels.ndim != 1 or len(number_of_pixels) != 2:
        raise ValueError(textwrap.fill((
            "The input `number_of_pixels` must be a list-like "
            "with exactly two values, corresponding to the "
            "number of pixels in the x- and y-dimension. "
            f"Received parameter with shape {number_of_pixels.shape}."
        )))

    if (number_of_pixels < 2).any():
        raise ValueError(textwrap.fill((
            "The input `number_of_pixels` must set at least two "
            "pixels in each dimension (i.e. all elements in "
            "`number_of_elements` must be larger or equal to two). "
            f"Received `{number_of_pixels}`."
        )))

    radius = np.float32(radius)

    if xlim is None:
        xlim = np.array([
            positions[:, 0].min() - radius,
            positions[:, 0].max() + radius,
        ], dtype = np.float32)
    else:
        xlim = np.asarray(xlim, dtype = np.float32)

        if xlim.ndim != 1 or len(xlim) != 2 or xlim[0] >= xlim[1]:
            raise ValueError(textwrap.fill((
                "The input `xlim` parameter must be a list with exactly "
                "two values, corresponding to the minimum and maximum "
                "coordinates of the pixel space in the x-dimension. "
                f"Received parameter with shape {xlim.shape}."
            )))

    if ylim is None:
        ylim = np.array([
            positions[:, 1].min() - radius,
            positions[:, 1].max() + radius,
        ], dtype = np.float32)
    else:
        ylim = np.asarray(ylim, dtype = np.float32)

        if ylim.ndim != 1 or len(ylim) != 2 or ylim[0] >= ylim[1]:
            raise ValueError(textwrap.fill((
                "The input `ylim` parameter must be a list with exactly "
                "two values, corresponding to the minimum and maximum "
                "coordinates of the pixel space in the y-dimension. "
                f"Received parameter with shape {ylim.shape}."
            )))

    omit_last = bool(omit_last)

    # `occupancy2d_ext` expects a single-precision pixels array
    grid = np.zeros(tuple(number_of_pixels), order = "C", dtype = np.float32)

    # This modifies `pixels` in-place
    occupancy2d_ext(
        grid,
        xlim,
        ylim,
        positions,
        times,
        radius,
        omit_last,
    )

    pixels = pept.Pixels(grid, xlim = xlim, ylim = ylim)

    if verbose:
        end = time.time()
        print(f"Computed occupancy grid in {end - start} s")

    return pixels



# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you must cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools


# File              : occupancy_ext.pyx
# License           : GNU v3.0
# Author            : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date              : 06.04.2020


# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: initializedcheck=False
# cython: nonecheck=False
# cython: embedsignature=True
# cython: cdivision=True


#import numpy as np

from libc.stdint cimport int32_t


cdef extern from "rtd2d.c":
    # C is included here so that it doesn't need to be compiled externally
    pass


cdef extern from "rtd2d.h":
    void rtd2d_occupancy(
        float           *pixels,
        const int32_t   dims[2],
        const float     xmax[2],
        const float     ymax[2],
        const float     *positions,
        const float     *times,
        const int32_t   num_positions,
        const float     radius,
        const int32_t   omit_first
    ) nogil


cpdef void occupancy2d_ext(
    float[:, :] pixels,
    const float[:] xlim,
    const float[:] ylim,
    const float[:, :] positions,
    const float[:] times,
    const float radius,
    bint omit_last = False,
) nogil:
    '''Compute the 2D occupancy of a single circular particle moving along a
    trajectory.

    ::

        Function signature:

            void occupancy2d_ext(
                float[:, :] pixels,
                float[:] xlim,
                float[:] ylim,
                float[:, :] positions,
                float[:] times,
                float radius,
                bool omit_last = False,
            ) nogil

    This corresponds to the pixellisation of moving circular particles, such
    that for every two consecutive particle locations, a 2D cylinder (i.e.
    convex hull of two circles at the two particle positions), the fraction of
    its area that intersets a pixel is multiplied with the time between the
    two particle locations and saved in the input `pixels`.

    All parameters sent to this function should be initialised properly. Please
    see the parameters' description below, or the function *will* segfault.
    Alternatively, use the `pept.processing.occupancy2d` function which is
    more robust and initialises all parameters for you.

    Parameters
    ----------
    pixels: (M, N) numpy.ndarray[ndim = 2, dtype = numpy.float32]
        The 2D grid of pixels, initialised to zero. It can be created with
        `numpy.zeros((nrows, ncols))`. Important that it contains
        single-precision values.

    xlim: np.float32
        The upper physical limit of the system spanned by `pixels` in the
        x-dimension. Important: the lower limit is always assumed to be at
        (0, 0). You can translate the particle positions before calling this
        function (`pept.processing.occupancy2d does this automatically`).

    ylim: np.float32
        The upper physical limit of the system spanned by `pixels` in the
        y-dimension. Important: the lower limit is always assumed to be at
        (0, 0). You can translate the particle positions before calling this
        function (`pept.processing.occupancy2d does this automatically`).

    positions: (P, 2) numpy.ndarray[ndim = 2, dtype = numpy.float32]
        The particles' 2D positions, where all rows are formatted as
        `[x_coordinate, y_coordinate]`.

    times: (P,) numpy.ndarray[ndim = 1, dtype = numpy.float32]
        The corresponding timestamp for each particle location in `positions`.

    radius: numpy.float32
        The particle's radius. Can be in any system of units, as long as it is
        consistent with what is used for the particle `positions`.

    omit_last: bool, default False
        If True, omit the last circle in the particle positions. Useful if
        rasterizing the same trajectory piece-wise; if you split the trajectory
        and call this function multiple times, set `omit_last = 0` to avoid
        considering the last particle location twice.

    '''

    cdef int32_t    dims[2]
    cdef int32_t    num_positions
    cdef int32_t    omit_last_i32 = omit_last

    dims[0] = pixels.shape[0]
    dims[1] = pixels.shape[1]

    num_positions = positions.shape[0]

    rtd2d_occupancy(
        &pixels[0, 0],
        dims,
        &xlim[0],
        &ylim[0],
        &positions[0, 0],
        &times[0],
        num_positions,
        radius,
        omit_last_i32
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you must cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools


# File      : birmingham_method.pyx
# License   : GNU v3.0
# Author    : Sam Manger
# Date      : 21.08.2019


# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: initializedcheck=False
# cython: nonecheck=False
# cython: embedsignature=True
# cython: cdivision=True


import numpy as np      # import numpy for Python functions
cimport numpy as np     # import numpy for C functions (numpy's C API)


cdef extern from "birmingham_method_ext.c":
    # C is included here so that it doesn't need to be compiled externally
    pass


cdef extern from "birmingham_method_ext.h":
    void birmingham_method_ext(
        const double *, const Py_ssize_t, const Py_ssize_t,
        double *, int *, const double
    ) nogil

    void calculate(
        double *, double *, double *, double *, double *, double *,
        double *, double *, double *, double *, double *, double *,
        double *, double *, double *, double *, double *,
        int *, int, int, double *
    ) nogil


# cpdef means it is defined both for Python and C code;
# cdef means it is defined only for C code;

# Cython has a cool function to automatically get memoryviews of the input
# parameters => double[:, :] receives a 2D numpy array.
cpdef birmingham_method(
    double[:, :] lines,
    double fopt
):
    '''Use the Birmingham Method to find one tracer location from the LoRs
    stored in `lines`.

    Function signature:
        birmingham_method(
            double[:, :] lines,     # LoRs in a sample
            double fopt             # Fraction of LoRs used to find tracer
        )

    This function receives a numpy array of LoRs (one "sample") from python,
    computing the minimum distance point (MDP). A number of lines that lie
    outside the standard deviation of the MDP are then removed from the set,
    and the MDP is recalculated. This process is repeated until approximately
    a fixed fraction (fopt) of the original lines is left.

    The found tracer position is then returned along with a boolean mask of
    the LoRs that were used to compute it.

    Parameters
    ----------
    lines : (N, M >= 7) numpy.ndarray
        A numpy array of the lines of respones (LoRs) that will be used to find
        a tracer location; each LoR is stored as a timestamp, the 3D
        coordinates of two points defining the line, followed by any additional
        data. The data columns are then `[time, x1, y1, z1, x2, y2, z2, etc]`.
        Note that the extra data is simply ignored by this function.
    fopt : float
        A float number between 0 and 1 representing the fraction of LoRs that
        will be used to compute the tracer location.

    Returns
    -------
    location : (5,) numpy.ndarray
        The computed tracer location, with data columns formatted as
        `[time, x, y, z, error]`.
    used : (N,) numpy.ndarray
        A boolean mask of the LoRs that were used to compute the tracer
        location; that is, a vector of the same length as `lines`, containing 1
        for the rows that were used, and 0 otherwise.

    Notes
    -----
    This is a low-level Cython function that does not do any checks on the
    input data - it is meant to be used in other modules / libraries. For a
    normal user, the `pept.tracking.birmingham_method.BirminghamMethod` class
    methods `fit_sample` and `fit` are recommended as higher-level APIs. They
    do check the input parameters and are easier to use.
    '''

    # Py_ssize_t is the one "strange" type from Cython - it is the type used
    # for indexing arrays, as pointers must have a certain number of bits that
    # is platform-dependent; that type is stored as a C macro in `ssize_t` that
    # Cython also provides as `Py_ssize_t`.
    cdef Py_ssize_t nrows = lines.shape[0]
    cdef Py_ssize_t ncols = lines.shape[1]

    # np.float64 == C double ; np.intc == C int ;
    cdef np.ndarray[double, ndim = 1] location = np.zeros(5, dtype = np.float64)
    cdef np.ndarray[int, ndim = 1] used = np.ones(nrows, dtype = np.intc)

    # Release the GIL as we're in a thread-safe C function for most of our
    # computation time.
    with nogil:
        birmingham_method_ext(
            &lines[0, 0],
            nrows,
            ncols,
            &location[0],
            &used[0],
            fopt
        )

    return location, used





#!/usr/bin/env python3
# -*- coding: utf-8 -*-


# File   : __init__.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 22.06.2020


'''The PEPT-oriented post-processing suite, including occupancy grid,
vector velocity fields, etc.

Summary
-------
This module contains fast, robust functions that operate on PEPT-like data
and integrate with the `pept` library's base classes.

Modules Provided
----------------

::

    pept.processing
    │
    Functions imported into the subpackage root:
    ├── circles2d :             Pixellise circles onto occupancy grid.
    ├── occupancy2d :           Pixellise trajectories onto occupancy grid.
    ├── circles2d_ext :         Circle pixellisation low-level Cython routine.
    └── occupancy2d_ext :       Trajectory pixellisation low-level C routine.

'''


from    .occupancy      import  circles2d
from    .occupancy      import  occupancy2d

from    .circles_ext    import  circles2d_ext
from    .occupancy_ext  import  occupancy2d_ext



__all__ = [
    "circles2d",
    "occupancy2d",
    "circles2d_ext",
    "occupancy2d_ext",
]


__license__ = "GNU v3.0"
__maintainer__ = "Andrei Leonard Nicusan"
__email__ = "a.l.nicusan@bham.ac.uk"
__status__ = "Beta"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools
#
#    Copyright (C) 2019 Andrei Leonard Nicusan
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.


# File              : get_pept_event.pyx
# License           : License: GNU v3.0
# Author            : Sam Manger <s.manger@bham.ac.uk>
# Date              : 27.06.2019


#!python
#cython: language_level=3


cdef extern from "get_pept_event_ext.c":
    # C is included here so that it doesn't need to be compiled externally
    pass

cdef extern from "get_pept_event_ext.h":
    void get_pept_event_ext(double *, unsigned int, int, int)
    void get_pept_LOR_ext(double *, unsigned int, int, int)

import numpy as np

def get_pept_event(word, itag, itime):
    # Lines for a single sample => n x 12 array
    # sampleLines row: [word time itag MPnum Bucket1 Bucket2 Block1 Block2 Seg1 Seg2 Plane1 Plane2]

    cdef unsigned int word_C = word
    cdef int itag_C = itag
    cdef int itime_C = itime

    data_array = np.zeros(12, order='C')      # Allocate enough memory
    # data_array = np.ravel(data_array, order='C')              # Cast into 1D array to send to C

    cdef double[::1] data_array_memview = data_array

    get_pept_event_ext(&data_array_memview[0], word_C, itag_C, itime_C)

    return data_array

def get_pept_LOR(word, itag, itime):
    # Lines for a single sample => n x 8 array
    # sampleLines row: [itag time X1 Y1 Z1 X2 Y2 Z2]

    cdef unsigned int word_C = word
    cdef int itag_C = itag
    cdef int itime_C = itime

    LOR = np.zeros(8, order='C')      # Allocate enough memory
    # data_array = np.ravel(data_array, order='C')              # Cast into 1D array to send to C

    cdef double[::1] LOR_memview = LOR

    get_pept_LOR_ext(&LOR_memview[0], word_C, itag_C, itime_C)

    return LOR



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


# File   : __init__.py
# License: License: GNU v3.0
# Author : Sam Manger <s.manger@bham.ac.uk>
# Date   : 20.08.2019


from    .modular_camera    import  ModularCamera


__all__ = [
    'ModularCamera'
]


__author__ = "Sam Manger"
__credits__ = [
    "Andrei Leonard Nicusan",
    "Kit Windows-Yule",
    "Sam Manger"
]
__license__ = "GNU v3.0"
__maintainer__ = "Sam Manger"
__email__ = "s.manger@bham.ac.uk"
__status__ = "Development"





#!/usr/bin/env python3
# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you must cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.


# File   : minpoints.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 20.10.2020


import  time
import  os
import  warnings
import  textwrap

import  numpy               as      np

from    joblib              import  Parallel, delayed
from    tqdm                import  tqdm

import  pept
from    pept.tracking       import  peptml


def find_minpoints(
    sample_lines,
    num_lines,
    max_distance,
    cutoffs = None,
    append_indices = False
):
    '''Compute the minimum distance points (MDPs) from all combinations of
    `num_lines` lines given in an array of lines `sample_lines`.

    Given a sample of lines, this functions computes the minimum distance
    points (MDPs) for every possible combination of `num_lines` lines. The
    returned numpy array contains all MDPs that satisfy the following:

    1. Are within the `cutoffs`.
    2. Are closer to all the constituent LoRs than `max_distance`.

    Parameters
    ----------
    sample_lines: (M, N) numpy.ndarray
        A 2D array of lines, where each line is defined by two points such that
        every row is formatted as `[t, x1, y1, z1, x2, y2, z2, etc.]`. It
        *must* have at least 2 lines and the combination size `num_lines`
        *must* be smaller or equal to the number of lines. Put differently:
        2 <= num_lines <= len(sample_lines).

    num_lines: int
        The number of lines in each combination of LoRs used to compute the
        MDP. This function considers every combination of `numlines` from the
        input `sample_lines`. It must be smaller or equal to the number of
        input lines `sample_lines`.

    max_distance: float
        The maximum allowed distance between an MDP and its constituent lines.
        If any distance from the MDP to one of its lines is larger than
        `max_distance`, the MDP is thrown away.

    cutoffs: (6,) numpy.ndarray, optional
        An array of spatial cutoff coordinates with *exactly 6 elements* as
        [x_min, x_max, y_min, y_max, z_min, z_max]. If any MDP lies outside
        this region, it is thrown away. If it is `None`, they are computed
        automatically by calling `get_cutoffs`. The default is `None`.

    append_indices: bool, default False
        A boolean specifying whether to include the indices of the lines used
        to compute each MDP. If `False`, the output array will only contain the
        [time, x, y, z] of the MDPs. If `True`, the output array will have
        extra columns [time, x, y, z, line_idx(1), ..., line_idx(n)] where
        n = `num_lines`.

    Returns
    -------
    minpoints: (M, N) numpy.ndarray
        A 2D array of `float`s containing the time and coordinates of the MDPs
        [time, x, y, z]. The time is computed as the average of the constituent
        lines. If `append_indices` is `True`, then `num_lines` indices of the
        constituent lines are appended as extra columns:
        [time, x, y, z, line_idx1, line_idx2, ..]. The first column (for time)
        is sorted.

    Raises
    ------
    ValueError
        If `sample_lines` is not a numpy array with shape (N, M >= 7).

    ValueError
        If 2 <= num_lines <= len(sample_lines) is not satisfied.

    ValueError
        If `cutoffs` is not a one-dimensional array with values
        `[min_x, max_x, min_y, max_y, min_z, max_z]`

    See Also
    --------
    pept.tracking.peptml.Minpoints : Compute minpoints from `pept.LineData`.
    pept.utilities.read_csv : Fast CSV file reading into numpy arrays.
    '''

    sample_lines = np.asarray(sample_lines, order = 'C', dtype = float)
    num_lines = int(num_lines)
    max_distance = float(max_distance)

    # Check sample has shape (N, M >= 7)
    if sample_lines.ndim != 2 or sample_lines.shape[1] < 7:
        raise ValueError((
            "\n[ERROR]: `sample_lines` should have dimensions (M, N), "
            f" where N >= 7. Received {sample_lines.shape}.\n"
        ))

    if not 2 <= num_lines <= len(sample_lines):
        raise ValueError((
            "\n[ERROR]: The number of lines in a combination must be smaller "
            "than the number of lines in the input `sample_lines`:\n"
            "2 <= num_lines <= len(sample_lines)\n"
        ))

    if cutoffs is None:
        cutoffs = peptml.get_cutoffs(sample_lines)
    else:
        cutoffs = np.asarray(cutoffs, order = 'C', dtype = float)
        if cutoffs.ndim != 1 or len(cutoffs) != 6:
            raise ValueError((
                "\n[ERROR]: cutoffs should be a one-dimensional array with "
                "values [min_x, max_x, min_y, max_y, min_z, max_z]. Received "
                f"{cutoffs}.\n"
            ))

    sample_minpoints = pept.utilities.find_minpoints(
        sample_lines,
        num_lines,
        max_distance,
        cutoffs,
        append_indices = append_indices
    )

    return sample_minpoints




class Minpoints(pept.PointData):
    '''Transform LoRs (a `pept.LineData` instance) into *minpoints* (a
    `pept.PointData` instance) for clustering, in parallel.

    Given a sample of lines, the minpoints are the minimum distance points
    (MDPs) for every possible combination of `num_lines` lines that satisfy
    the following conditions:

    1. Are within the `cutoffs`.
    2. Are closer to all the constituent LoRs than `max_distance`.

    Under typical usage, the `Minpoints` class is initialised with a
    `pept.LineData` instance, automatically calculating the minpoints from the
    samples of lines. The `Minpoints` class inherits from `pept.PointData`,
    such that once the cutpoints have been computed, all the methods from the
    parent class `pept.PointData` can be used on them (such as visualisation
    functionality).

    For more control over the operations, `pept.tracking.peptml.find_minpoints`
    can be used - it receives a generic numpy array of LoRs (one 'sample') and
    returns a numpy array of cutpoints.

    Attributes
    ----------
    line_data : instance of pept.LineData
        The LoRs for which the cutpoints will be computed. It must be an
        instance of `pept.LineData`.

    num_lines: int
        The number of lines in each combination of LoRs used to compute the
        MDP. This function considers every combination of `num_lines` from the
        input `sample_lines`. It must be smaller or equal to the number of
        input lines `sample_lines`.

    max_distance: float
        The maximum allowed distance between an MDP and its constituent lines.
        If any distance from the MDP to one of its lines is larger than
        `max_distance`, the MDP is thrown away. A good starting value would be
        0.1 mm for small tracers and/or clean data, or 0.2 mm for larger
        tracers and/or noisy data.

    cutoffs : list-like of length 6
        A list (or equivalent) of the cutoff distances for every axis,
        formatted as `[x_min, x_max, y_min, y_max, z_min, z_max]`. Only the
        minpoints which fall within these cutoff distances are considered. The
        default is None, in which case they are automatically computed using
        `pept.tracking.peptml.get_cutoffs`.

    sample_size, overlap, number_of_lines, etc. : inherited from pept.PointData
        Additional attributes and methods are inherited from the base class
        `PointData`. Check its documentation for more information.

    Methods
    -------
    find_minpoints(line_data, num_lines, max_distance, cutoffs = None,\
                   append_indices = False, max_workers = None, verbose = True)
        Compute the minpoints from the samples in a `LineData` instance.

    sample, to_csv, plot, etc. : inherited from pept.PointData
        Additional attributes and methods are inherited from the base class
        `PointData`. Check its documentation for more information.

    Notes
    -----
    Once instantiated with a `LineData`, the class computes the minpoints and
    *automatically sets the sample_size* to the average number of minpoints
    found per sample of LoRs.

    Examples
    --------
    Compute the minpoints for a `LineData` instance for all triplets of lines
    that are less than 0.1 from those lines:

    >>> line_data = pept.LineData(example_data)
    >>> minpts = peptml.Minpoints(line_data, 3, 0.1)

    Compute the minpoints for a single sample:

    >>> sample = line_data[0]
    >>> cutpts_sample = peptml.find_minpoints(sample, 3, 0.1)

    See Also
    --------
    pept.LineData : Encapsulate LoRs for ease of iteration and plotting.
    pept.tracking.peptml.HDBSCANClusterer : Efficient, parallel HDBSCAN-based
                                            clustering of cutpoints.
    pept.scanners.ParallelScreens : Read in and initialise a `pept.LineData`
                                    instance from parallel screens PET/PEPT
                                    detectors.
    pept.utilities.read_csv : Fast CSV file reading into numpy arrays.
    '''

    def __init__(
        self,
        line_data,
        num_lines,
        max_distance,
        cutoffs = None,
        append_indices = False,
        max_workers = None,
        verbose = True
    ):
        '''Cutpoints class constructor.

        Parameters
        ----------
        line_data : instance of pept.LineData
            The LoRs for which the cutpoints will be computed. It must be an
            instance of `pept.LineData`.

        num_lines: int
            The number of lines in each combination of LoRs used to compute the
            MDP. This function considers every combination of `num_lines` from
            the input `sample_lines`. It must be smaller or equal to the number
            of input lines `sample_lines`.

        max_distance: float
            The maximum allowed distance between an MDP and its constituent
            lines. If any distance from the MDP to one of its lines is larger
            than `max_distance`, the MDP is thrown away. A good starting value
            would be 0.1 mm for small tracers and/or clean data, or 0.2 mm for
            larger tracers and/or noisy data.


        cutoffs : list-like of length 6, optional
            A list (or equivalent) of the cutoff distances for every axis,
            formatted as `[x_min, x_max, y_min, y_max, z_min, z_max]`. Only the
            minpoints which fall within these cutoff distances are considered.
            The default is None, in which case they are automatically computed
            using `pept.tracking.peptml.get_cutoffs`.

        append_indices : bool, default False
            If set to `True`, the indices of the individual LoRs that were used
            to compute each minpoint are also appended to the returned array.

        max_workers : int, optional
            The maximum number of threads that will be used for asynchronously
            computing the minpoints from the samples of LoRs in `line_data`.

        verbose : bool, default True
            Provide extra information when computing the cutpoints: time the
            operation and show a progress bar.

        Raises
        ------
        TypeError
            If `line_data` is not an instance of `pept.LineData`.

        ValueError
            If 2 <= num_lines <= len(sample_lines) is not satisfied.

        ValueError
            If `cutoffs` is not a one-dimensional array with values formatted
            as `[min_x, max_x, min_y, max_y, min_z, max_z]`.
        '''

        # Find the cutpoints when instantiated. The method
        # also initialises the instance as a `PointData` subclass.
        self.find_minpoints(
            line_data,
            num_lines,
            max_distance,
            cutoffs = cutoffs,
            append_indices = append_indices,
            max_workers = max_workers,
            verbose = verbose
        )


    @property
    def line_data(self):
        return self._line_data


    @property
    def num_lines(self):
        return self._num_lines


    @property
    def max_distance(self):
        return self._max_distance


    @property
    def cutoffs(self):
        return self._cutoffs


    def find_minpoints(
        self,
        line_data,
        num_lines,
        max_distance,
        cutoffs = None,
        append_indices = False,
        max_workers = None,
        verbose = True
    ):
        '''Compute the cutpoints from the samples in a `LineData` instance.

        Parameters
        ----------
        line_data : instance of pept.LineData
            The LoRs for which the cutpoints will be computed. It must be an
            instance of `pept.LineData`.

        num_lines: int
            The number of lines in each combination of LoRs used to compute the
            MDP. This function considers every combination of `num_lines` from
            the input `sample_lines`. It must be smaller or equal to the number
            of input lines `sample_lines`.

        max_distance: float
            The maximum allowed distance between an MDP and its constituent
            lines. If any distance from the MDP to one of its lines is larger
            than `max_distance`, the MDP is thrown away. A good starting value
            would be 0.1 mm for small tracers and/or clean data, or 0.2 mm for
            larger tracers and/or noisy data.


        cutoffs : list-like of length 6, optional
            A list (or equivalent) of the cutoff distances for every axis,
            formatted as `[x_min, x_max, y_min, y_max, z_min, z_max]`. Only the
            minpoints which fall within these cutoff distances are considered.
            The default is None, in which case they are automatically computed
            using `pept.tracking.peptml.get_cutoffs`.

        append_indices : bool, default False
            If set to `True`, the indices of the individual LoRs that were used
            to compute each minpoint are also appended to the returned array.

        max_workers : int, optional
            The maximum number of threads that will be used for asynchronously
            computing the minpoints from the samples of LoRs in `line_data`.

        verbose : bool, default True
            Provide extra information when computing the cutpoints: time the
            operation and show a progress bar.

        Returns
        -------
        self : the PointData instance of cutpoints
            The computed cutpoints are stored in the `Cutpoints` class, as a
            subclass of `pept.PointData`.

        Raises
        ------
        TypeError
            If `line_data` is not an instance of `pept.LineData`.

        ValueError
            If 2 <= num_lines <= len(sample_lines) is not satisfied.

        ValueError
            If `cutoffs` is not a one-dimensional array with values formatted
            as `[min_x, max_x, min_y, max_y, min_z, max_z]`.

        Notes
        -----
        This method is automatically called when instantiating this class.
        '''

        if verbose:
            start = time.time()

        # Check line_data is an instance (or a subclass!) of pept.LineData
        if not isinstance(line_data, pept.LineData):
            raise TypeError((
                "\n[ERROR]: line_data should be an instance (or subclass) of "
                "`pept.LineData`.\n"
            ))

        if not 2 <= num_lines <= line_data.sample_size:
            raise ValueError((
                "\n[ERROR]: The number of lines in a combination must be "
                "smaller than the LineData sample_size:"
                "\n2 <= num_lines <= line_data.sample_size\n"
            ))

        # Users might forget to set the sample_size, leaving it to the default
        # value of 0; in that case, all lines are returned as a single sample -
        # that might not be the intended behaviour.
        if line_data.sample_size == 0:
            warnings.warn(
                textwrap.fill((
                    "\n[WARNING]: The `line_data.sample_size` was left to the "
                    "default value of 0, in which case all lines are returned "
                    "as a single sample. For a very large number of lines, "
                    "this might result in a long function execution time.\n"
                ), replace_whitespace = False),
                RuntimeWarning
            )

        self._line_data = line_data
        self._max_distance = float(max_distance)
        self._num_lines = int(num_lines)

        # If cutoffs were not supplied, compute them
        if cutoffs is None:
            cutoffs = peptml.get_cutoffs(line_data.lines)
        # Otherwise make sure they are a C-contiguous numpy array
        else:
            cutoffs = np.asarray(cutoffs, order = 'C', dtype = float)
            if cutoffs.ndim != 1 or len(cutoffs) != 6:
                raise ValueError((
                    "\n[ERROR]: cutoffs should be a one-dimensional array "
                    "with values [min_x, max_x, min_y, max_y, min_z, max_z]. "
                    f"Received {cutoffs}.\n"
                ))

        self._cutoffs = cutoffs

        # Using ThreadPoolExecutor, asynchronously collect the minpoints from
        # every sample in a list of arrays. This is more efficient than using
        # ProcessPoolExecutor because find_minpoints is a Cython function that
        # releases the GIL for most of its computation.
        # If verbose, show progress bar using tqdm.
        if max_workers is None:
            max_workers = os.cpu_count()

        '''
        with ThreadPoolExecutor(max_workers = max_workers) as executor:
            futures = []
            for sample in line_data:
                futures.append(
                    executor.submit(
                        pept.utilities.find_minpoints,
                        sample,
                        self._num_lines,
                        self._max_distance,
                        self._cutoffs,
                        append_indices = append_indices
                    )
                )

            if verbose:
                futures = tqdm(futures)

            minpoints = [f.result() for f in futures]
        '''

        minpoints = Parallel(n_jobs = max_workers)(
            delayed(pept.utilities.find_minpoints)(
                sample,
                self._num_lines,
                self._max_distance,
                self._cutoffs,
                append_indices = append_indices
            ) for sample in tqdm(line_data)
        )


        # minpoints shape: (n, m, 4), where n is the number of samples, and
        # m is the number of minpoints in the sample
        number_of_samples = len(minpoints)
        minpoints = np.vstack(minpoints)
        number_of_minpoints = len(minpoints)

        # Average number of minpoints per sample
        minpoints_per_sample = int(number_of_minpoints / number_of_samples)

        pept.PointData.__init__(
            self,
            minpoints,
            sample_size = minpoints_per_sample,
            overlap = 0,
            verbose = False
        )

        if verbose:
            end = time.time()
            print(f"\nFinding the minpoints took {end - start} seconds.\n")

        return self


    def __repr__(self):
        # Called when writing the class on a REPL. Add another line to the
        # standard description given in the parent class, pept.PointData.
        docstr = pept.PointData.__repr__(self) + (
            "\n\nNotes\n-----\n"
            "Once instantiated with a `LineData`, the class computes the \n"
            "minpoints and *automatically sets the sample_size* to the \n"
            "average number of minpoints found per sample of LoRs."
        )

        return docstr



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : read_csv.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 14.04.2020


import pandas as pd
import csv


def number_of_lines(filepath_or_buffer):
    '''Return the number of lines (or rows) in a file.

    Parameters
    ----------
    filepath_or_buffer : str, path object or file-like object
        Path to the file.

    Returns
    -------
    int
        The number of lines in the file pointed at by `filepath_or_buffer`.
    '''

    with open(filepath_or_buffer) as f:
        file_lines = sum(1 for line in f)

    return file_lines


def read_csv(
    filepath_or_buffer,         # Essential
    skiprows = None,            # Important
    nrows = None,               # Important
    dtype = float,              # Medium Importance
    sep = "\s+",                # Extra parameters
    header = None,              #       |
    engine = "c",               #       |
    na_filter = False,          #       |
    quoting = csv.QUOTE_NONE,   #       |
    memory_map = True,          #       -
    **kwargs                    # Extra keyword arguments to pandas.read_csv
):
    '''Read a given number of lines from a file and return a numpy array of the
    values.

    This is a convenience function that's simply a proxy to `pandas.read_csv`,
    configured with default parameters for fast reading and parsing of usual
    PEPT data.

    Most importantly, it reads from a **space-separated values** file at
    `filepath_or_buffer`, optionally skipping `skiprows` lines and reading in
    `nrows` lines. It returns a `numpy.ndarray` with `float` values.

    The parameters below are sent to `pandas.read_csv` with no further parsing.
    The descriptions below are taken from the `pandas` documentation.

    Parameters
    ----------
    filepath_or_buffer : str, path object or file-like object
        Any valid string path is acceptable. The string could be a URL. Valid
        URL schemes include http, ftp, s3, and file. For file URLs, a host is
        expected. A local file could be: file://localhost/path/to/table.csv. If
        you want to pass in a path object, pandas accepts any `os.PathLike`. By
        file-like object, we refer to objects with a `read()` method, such as a
        file handler (e.g. via builtin `open` function) or `StringIO`.
    skiprows : list-like, int or callable, optional
        Line numbers to skip (0-indexed) or number of lines to skip (int) at
        the start of the file.
    nrows : int, optional
        Number of rows of file to read. Useful for reading pieces of large
        files.
    dtype : Type name, default `float`
        Data type for data or columns. E.g. {‘a’: np.float64, ‘b’: np.int32,
        ‘c’: ‘Int64’}.
    sep : str, default `"\s+"`
        Delimiter to use. Separators longer than 1 character and different from
        '\s+' will be interpreted as regular expressions and will also force
        the use of the Python parsing engine.
    header : int, list of int, "infer", optional
        Row number(s) to use as the column names, and the start of the data. By
        default assume there is no header present (i.e. `header = None`).
    engine : {‘c’, ‘python’}, default "c"
        Parser engine to use. The C engine is faster while the python engine is
        currently more feature-complete.
    na_filter : bool, default `True`
        Detect missing value markers (empty strings and the value of
        na_values). In data without any NAs, passing na_filter=False can
        improve the performance of reading a large file.
    quoting : int or csv.QUOTE_* instance, default `csv.QUOTE_NONE`
        Control field quoting behavior per csv.QUOTE_* constants. Use one of
        QUOTE_MINIMAL (0), QUOTE_ALL (1), QUOTE_NONNUMERIC (2) or
        QUOTE_NONE (3).
    memory_map : bool, default True
        If a filepath is provided for filepath_or_buffer, map the file object
        directly onto memory and access the data directly from there. Using
        this option can improve performance because there is no longer any I/O
        overhead.
    kwargs : optional
        Extra keyword arguments that will be passed to `pandas.read_csv`.
    '''

    data = pd.read_csv(
        filepath_or_buffer,
        skiprows = skiprows,
        nrows = nrows,
        dtype = dtype,
        sep = sep,
        header = header,
        engine = engine,
        na_filter = na_filter,
        quoting = quoting,
        memory_map = memory_map,
        **kwargs
    )

    return data.to_numpy()


def read_csv_chunks(
    filepath_or_buffer,
    chunksize,
    skiprows = None,
    nrows = None,
    dtype = float,
    sep = "\s+",
    header = None,
    engine = "c",
    na_filter = False,
    quoting = csv.QUOTE_NONE,
    memory_map = True,
    **kwargs
):
    '''Read chunks of data from a file lazily, returning numpy arrays of the
    values.

    This function returns a generator - an object that can be iterated over
    once, creating data on-demand. This means that chunks of data will be
    read only when being accessed, making it a more efficient alternative to
    `read_csv` for large files (> 1.000.000 lines).

    A more convenient and feature-complete alternative is
    `pept.utilities.ChunkReader` which is more reusable and can access
    out-of-order chunks using subscript notation (i.e. data[0]).

    This is a convenience function that's simply a proxy to `pandas.read_csv`,
    configured with default parameters for fast reading and parsing of usual
    PEPT data.

    Most importantly, it lazily read chunks of size `chunksize` from a
    **space-separated values** file at `filepath_or_buffer`, optionally
    skipping `skiprows` lines and reading in `nrows` lines. It returns
    `numpy.ndarray`s with `float` values.

    The parameters below are sent to `pandas.read_csv` with no further parsing.
    The descriptions below are taken from the `pandas` documentation.

    Parameters
    ----------
    filepath_or_buffer : str, path object or file-like object
        Any valid string path is acceptable. The string could be a URL. Valid
        URL schemes include http, ftp, s3, and file. For file URLs, a host is
        expected. A local file could be: file://localhost/path/to/table.csv. If
        you want to pass in a path object, pandas accepts any `os.PathLike`. By
        file-like object, we refer to objects with a `read()` method, such as a
        file handler (e.g. via builtin `open` function) or `StringIO`.
    chunksize : int
        Number of lines read in a chunk of data. Return TextFileReader object
        for iteration.
    skiprows : list-like, int or callable, optional
        Line numbers to skip (0-indexed) or number of lines to skip (int) at
        the start of the file.
    nrows : int, optional
        Number of rows of file to read. Useful for reading pieces of large
        files.
    dtype : Type name, default `float`
        Data type for data or columns. E.g. {‘a’: np.float64, ‘b’: np.int32,
        ‘c’: ‘Int64’}.
    sep : str, default `"\s+"`
        Delimiter to use. Separators longer than 1 character and different from
        '\s+' will be interpreted as regular expressions and will also force
        the use of the Python parsing engine.
    header : int, list of int, "infer", optional
        Row number(s) to use as the column names, and the start of the data. By
        default assume there is no header present (i.e. `header = None`).
    engine : {‘c’, ‘python’}, default "c"
        Parser engine to use. The C engine is faster while the python engine is
        currently more feature-complete.
    na_filter : bool, default `True`
        Detect missing value markers (empty strings and the value of
        na_values). In data without any NAs, passing na_filter=False can
        improve the performance of reading a large file.
    quoting : int or csv.QUOTE_* instance, default `csv.QUOTE_NONE`
        Control field quoting behavior per csv.QUOTE_* constants. Use one of
        QUOTE_MINIMAL (0), QUOTE_ALL (1), QUOTE_NONNUMERIC (2) or
        QUOTE_NONE (3).
    memory_map : bool, default True
        If a filepath is provided for filepath_or_buffer, map the file object
        directly onto memory and access the data directly from there. Using
        this option can improve performance because there is no longer any I/O
        overhead.
    kwargs : optional
        Extra keyword arguments that will be passed to `pandas.read_csv`.
    '''

    reader = pd.read_csv(
        filepath_or_buffer,
        chunksize = chunksize,
        skiprows = skiprows,
        nrows = nrows,
        dtype = dtype,
        sep = sep,
        header = header,
        engine = engine,
        na_filter = na_filter,
        quoting = quoting,
        memory_map = memory_map,
        **kwargs
    )

    for chunk in reader:
        yield chunk.values




class ChunkReader:
    '''Class for fast, on-demand reading / parsing and iteration over chunks of
    data from CSV files.

    This is an abstraction above `pandas.read_csv` for easy and fast iteration
    over chunks of data from a CSV file. The chunks can be accessed using
    normal iteration (`for chunk in reader: ...`) and subscripting
    (`reader[0]`).

    The chunks are read lazily, only upon access. It is therefore a more
    efficient alternative to `read_csv` for large files (> 1.000.000 lines).
    For convenience, this class configures some default parameters for
    `pandas.read_csv` for fast reading and parsing of usual PEPT data.

    Most importantly, it reads chunks containing `chunksize` lines from a
    **space-separated values** file at `filepath_or_buffer`, optionally
    skipping `skiprows` lines and reading in at most `nrows` lines. It returns
    `numpy.ndarray`s with `float` values.

    Attributes
    ----------
    filepath_or_buffer : str, path object or file-like object
        Any valid string path is acceptable. The string could be a URL.
        Valid URL schemes include http, ftp, s3, and file. For file URLs, a
        host is expected. A local file could be
        file://localhost/path/to/table.csv. If you want to pass in a path
        object, pandas accepts any `os.PathLike`. By file-like object, we
        refer to objects with a `read()` method, such as a file handler
        (e.g. via builtin `open` function) or `StringIO`.
    number_of_chunks : int
        The number of chunks (also returned when using the `len` method),
        taking into account the lines skipped (`skiprows`), the number of lines
        in the file (`file_lines`) and the maximum number of lines to be read
        (`nrows`).
    file_lines : int
        The number of lines in the file pointed at by `filepath_or_buffer`.
    chunksize : int
        The number of lines in a chunk of data.
    skiprows : int
        The number of lines to be skipped at the beginning of the file.
    nrows : int
        The maximum number of lines to be read. Only has an effect if it is
        less than `file_lines` - `skiprows`. For example, if a file has
        10 lines and `skiprows` = 5 and `chunksize` = 5, even if `nrows` were
        to be 20, the `number_of_chunks` should still be 1.

    Raises
    ------
    IndexError
        Upon access to a non-existent chunk using subscript notation
        (i.e. `data[100]` when there are 50 chunks).

    Examples
    --------
    Say "data.csv" contains 1_000_000 lines of data. Read chunks of 10_000
    lines as a time, skipping the first 100_000:

    >>> from pept.utilities import ChunkReader
    >>> chunks = ChunkReader("data.csv", 10_000, skiprows = 100_000)
    >>> len(chunks)         # 90 chunks
    >>> chunks.file_lines   # 1_000_000

    Normal iteration:

    >>> for chunk in chunks:
    >>>     ... # neat operations

    Access a single chunk using subscripting:

    >>> chunks[0]   # First chunk
    >>> chunks[-1]  # Last chunk
    >>> chunks[100] # IndexError

    See Also
    --------
    pept.utilities.read_csv : Fast CSV file reading into numpy arrays.
    pept.LineData : Encapsulate LoRs for ease of iteration and plotting.
    pept.PointData : Encapsulate points for ease of iteration and plotting.
    '''

    def __init__(
        self,
        filepath_or_buffer,
        chunksize,
        skiprows = None,
        nrows = None,
        dtype = float,
        sep = "\s+",
        header = None,
        engine = "c",
        na_filter = False,
        quoting = csv.QUOTE_NONE,
        memory_map = True,
        **kwargs
    ):
        '''ChunkReader class constructor.

        Parameters
        ----------
        filepath_or_buffer : str, path object or file-like object
            Any valid string path *to a local file* is acceptable. If you want
            to read in lines from an online location (i.e. using a URL), you
            should use `pept.utilities.read_csv`. If you want to pass in a path
            object, pandas accepts any `os.PathLike`. By file-like object, we
            refer to objects with a `read()` method, such as a file handler
            (e.g. via builtin `open` function) or `StringIO`.
        chunksize : int
            Number of lines read in a chunk of data.
        skiprows : list-like, int or callable, optional
            Line numbers to skip (0-indexed) or number of lines to skip (int)
            at the start of the file.
        nrows : int, optional
            Number of rows of file to read. Useful for reading pieces of large
            files.
        dtype : Type name, default `float`
            Data type for data or columns. E.g. {‘a’: np.float64,
            ‘b’: np.int32, ‘c’: ‘Int64’}.
        sep : str, default `"\s+"`
            Delimiter to use. Separators longer than 1 character and different
            from '\s+' will be interpreted as regular expressions and will also
            force the use of the Python parsing engine.
        header : int, list of int, "infer", optional
            Row number(s) to use as the column names, and the start of the
            data. By default assume there is no header present (i.e.
            `header = None`).
        engine : {‘c’, ‘python’}, default "c"
            Parser engine to use. The C engine is faster while the python
            engine is currently more feature-complete.
        na_filter : bool, default `True`
            Detect missing value markers (empty strings and the value of
            na_values). In data without any NAs, passing na_filter=False can
            improve the performance of reading a large file.
        quoting : int or csv.QUOTE_* instance, default `csv.QUOTE_NONE`
            Control field quoting behavior per csv.QUOTE_* constants. Use one
            of QUOTE_MINIMAL (0), QUOTE_ALL (1), QUOTE_NONNUMERIC (2) or
            QUOTE_NONE (3).
        memory_map : bool, default True
            If a filepath is provided for filepath_or_buffer, map the file
            object directly onto memory and access the data directly from
            there. Using this option can improve performance because there is
            no longer any I/O overhead.
        kwargs : optional
            Extra keyword arguments that will be passed to `pandas.read_csv`.

        Raises
        ------
        EOFError : End Of File Error
            If `skiprows` >= `number_of_lines`.
        '''

        self.filepath_or_buffer = filepath_or_buffer
        self._chunksize = chunksize

        self._file_lines = number_of_lines(filepath_or_buffer)

        if skiprows is None:
            self._skiprows = 0
        elif skiprows >= self.file_lines:
            raise EOFError((
                f"\n[ERROR]: Tried to skip `skiprows` = {skiprows} lines "
                f"when there are `file_lines` = {self._file_lines} lines in "
                "the data file.\n"
            ))
        else:
            self._skiprows = skiprows

        # If undefined, set `nrows` to the maximum number of lines that can be
        # read from the file; that is `file_lines` - `skiprows`
        if nrows is None:
            self._nrows = self._file_lines - self._skiprows
        else:
            self._nrows = nrows

        self.dtype = dtype
        self.sep = sep
        self.header = header
        self.engine = engine
        self.na_filter = na_filter
        self.quoting = quoting
        self.memory_map = memory_map
        self.kwargs = kwargs

        # The number of chunks is (the number of lines that can be read from
        # the file OR the set number of rows to be read, whichever's smaller)
        # divided by the chunksize.
        self._number_of_chunks = int(
            min(self._nrows, self._file_lines - self._skiprows) / \
            self._chunksize
        )
        self._index = 0


    @property
    def number_of_chunks(self):
        return self._number_of_chunks


    @property
    def file_lines(self):
        return self._file_lines


    @property
    def chunksize(self):
        return self._chunksize


    @chunksize.setter
    def chunksize(self, new_chunksize):
        self._chunksize = new_chunksize

        # Recalculate the number of chunks and reset the inner index
        self._number_of_chunks = int(
            min(self._nrows, self._file_lines - self._skiprows) / \
            self._chunksize
        )
        self._index = 0


    @property
    def skiprows(self):
        return self._skiprows


    @skiprows.setter
    def skiprows(self, new_skiprows):
        if new_skiprows is None:
            self._skiprows = 0
        elif new_skiprows >= self._file_lines:
            raise EOFError((
                f"\n[ERROR]: Tried to skip `skiprows` = {new_skiprows} lines "
                f"when there are `file_lines` = {self._file_lines} lines in "
                "the data file.\n"
            ))
        else:
            self._skiprows = new_skiprows

        # Recalculate the number of chunks and reset the inner index
        self._number_of_chunks = int(
            min(self._nrows, self._file_lines - self._skiprows) / \
            self._chunksize
        )
        self._index = 0


    @property
    def nrows(self):
        return self._nrows


    @nrows.setter
    def nrows(self, new_nrows):
        if new_nrows is None:
            self._nrows = self._file_lines - self._skiprows
        else:
            self._nrows = new_nrows

        # Recalculate the number of chunks and reset the inner index
        self._number_of_chunks = int(
            min(self._nrows, self._file_lines - self._skiprows) / \
            self._chunksize
        )
        self._index = 0


    def __len__(self):
        return self._number_of_chunks


    def __iter__(self):
        return self


    def __next__(self):
        if self._index >= self._number_of_chunks:
            self._index = 0
            raise StopIteration

        data = pd.read_csv(
            self.filepath_or_buffer,
            skiprows = self._skiprows + self._index * self._chunksize,
            nrows = self._chunksize,
            dtype = self.dtype,
            sep = self.sep,
            header = self.header,
            engine = self.engine,
            na_filter = self.na_filter,
            quoting = self.quoting,
            memory_map = self.memory_map,
            **self.kwargs
        )

        self._index = self._index + 1

        return data.to_numpy()


    def __getitem__(self, key):
        if key >= self._number_of_chunks:
            raise IndexError((
                f"\n[ERROR]: Tried to read the data chunk at index {key} when "
                f"there are {self._number_of_chunks} chunks (indexed from "
                "0).\n"
            ))

        # Allow negative indices
        while key < 0:
            key += self._number_of_chunks

        data = pd.read_csv(
            self.filepath_or_buffer,
            skiprows = self._skiprows + key * self._chunksize,
            nrows = self._chunksize,
            dtype = self.dtype,
            sep = self.sep,
            header = self.header,
            engine = self.engine,
            na_filter = self.na_filter,
            quoting = self.quoting,
            memory_map = self.memory_map,
            **self.kwargs
        )

        return data.to_numpy()


    def __str__(self):
        # Shown when calling print(class)
        docstr = (
            f"filepath_or_buffer = {self.filepath_or_buffer}\n"
            f"file_lines =         {self.file_lines}\n\n"
            f"skiprows =           {self.skiprows}\n"
            f"nrows =              {self.nrows}\n\n"
            f"chunksize =          {self.chunksize}\n"
            f"number_of_chunks =   {self.number_of_chunks}"
        )

        return docstr


    def __repr__(self):
        # Shown when writing the class on a REPL
        docstr = (
            "Class instance that inherits from `pept.utilities.ChunkReader`.\n"
            f"Type:\n{type(self)}\n\n"
            "Attributes\n"
            "----------\n"
            f"{self.__str__()}\n"
        )

        return docstr
